import argparse
from collections import defaultdict
import logging
import pathlib
from typing import Iterable
import math
import copy

import yaml

import retinapy
import retinapy._logging
import retinapy.dataset
import retinapy.vis
import retinapy.mea as mea
import retinapy.models
import retinapy.train
import retinapy.nn
import scipy
import torch
import plotly.io
import torchinfo
import einops


DEFAULT_OUT_BASE_DIR = "./out/"
LOG_FILENAME = "train.log"
ARGS_FILENAME = "args.yaml"
TENSORBOARD_DIR = "tensorboard"

IN_CHANNELS = 4 + 1
# Quite often there are lengths in the range 300.
# The pad acts as the maximum, so it's a good candidate for a norm factor.
# Example: setting normalization to 400 would cause 400 time steps to be fit
# into the [0,1] region.
LOSS_CALC_PAD_MS = 600
DIST_CLAMP_MS = 600
SPLIT_RATIO = (7, 2, 1)

_logger = logging.getLogger(__name__)


def arg_parsers():

    """Parse commandline and config file arguments.

    The approach carried out here is inspired by the pytorch-image-models
    project:
        https://github.com/rwightman/pytorch-image-models

    Arguments are populated in the following order:
        1. Default values
        2. Config file
        3. Command line
    """
    config_parser = argparse.ArgumentParser(
        description="Config from YAML", add_help=False
    )
    config_parser.add_argument(
        "-c",
        "--config",
        default=None,
        type=str,
        metavar="FILE",
        help="YAML config file to override argument defaults.",
    )

    parser = argparse.ArgumentParser(description="Spike detection training")
    # fmt: off

    # Model/config arguments
    # Using -k as a filter, just like pytest.
    parser.add_argument("-k", type=str, default=None, metavar="EXPRESSION", help="Filter configs and models to train or test.")

    # Optimization parameters
    opt_group = parser.add_argument_group("Optimizer parameters")
    opt_group.add_argument("--lr", type=float, default=1e-5, help="Learning rate.")
    opt_group.add_argument("--weight-decay", type=float, default=1e-6, help="weight decay (default: 2e-5)")

    # Data
    data_group = parser.add_argument_group("Data parameters")
    data_group.add_argument("--data-dir", type=str, default=None, metavar="FILE", help="Path to stimulus pattern file.")
    data_group.add_argument("--recording-names", nargs='+', type=str, default=None, help="Names of recordings within the recording file.")
    data_group.add_argument("--cluster-ids", nargs='+', type=int, default=None, help="Cluster ID to train on.")

    parser.add_argument("--steps-til-eval", type=int, default=None, help="Steps until validation.")
    parser.add_argument("--steps-til-log", type=int, default=100, help="How many batches to wait before logging a status update.")
    parser.add_argument("--steps-til-eval-test-ds", type=int, default=10, help="After how many validation runs with the validation data should validation be run with the training data.")
    parser.add_argument("--initial-checkpoint", type=str, default=None, help="Initialize model from the checkpoint at this path.")
    #parser.add_argument("--resume", type=str, default=None, help="Resume full model and optimizer state from checkpoint path.")
    parser.add_argument("--output", type=str, default=None, metavar="DIR", help="Path to output folder (default: current dir).")
    parser.add_argument("--labels", type=str, default=None, help="List of experiment labels. Used for naming files and/or subfolders.")
    parser.add_argument("--epochs", type=int, default=8, metavar="N", help="number of epochs to train (default: 300)")
    parser.add_argument("--batch-size", type=int, default=128, help="batch size")
    parser.add_argument("--num-workers", type=int, default=24, help="Number of workers for data loading.")
    parser.add_argument("--pin-memory", action=argparse.BooleanOptionalAction, help="Pin memory?")

    model_group = parser.add_argument_group("Model options")
    model_group.add_argument("--zdim", type=int, default=2, help="VAE latent dimension")
    model_group.add_argument("--vae-beta", type=float, default=0.01, help="VAE beta parameter.")
    model_group.add_argument("--stride", type=int, default=17, help="Dataset stride.")
    model_group.add_argument("--num-heads", type=int, default=8, help="Number of transformer heads.")
    model_group.add_argument("--head-dim", type=int, default=32, help="Dimension of transformer heads.")
    model_group.add_argument("--num-tlayers", type=int, default=5, help="Number of transformer layers.")
    # fmt: on
    return config_parser, parser


def parse_args():
    config_parser, parser = arg_parsers()
    # First check if we have a config file to deal with.
    args, remaining = config_parser.parse_known_args()
    if args.config:
        with open(args.config, "r") as f:
            config = yaml.safe_load(f)
            # Populate the main parser's defaults.
            parser.set_defaults(**config)
    # Now the main parser.
    opt = parser.parse_args(remaining)
    # Serialize the arguments.
    opt_text = yaml.safe_dump(opt.__dict__, default_flow_style=False)
    return opt, opt_text


def args_from_yaml(yaml_path):
    _, parser = arg_parsers()
    with open(yaml_path, "r") as f:
        args = yaml.safe_load(f)
    parser.set_defaults(**args)
    opt = parser.parse_args(args=[])
    return opt


class Configuration:
    def __init__(self, downsample, input_len, output_len):
        self.downsample = downsample
        self.input_len = input_len
        self.output_len = output_len

    def __str__(self):
        return f"{self.downsample}ds_{self.input_len}in_{self.output_len}out"


"""
The following table is useful as a reference to see the effects of each 
downsampling factor. The downsampling rates in the table were chosen to fall
closely to some "simple" frequencies/periods, like 1 kHz, etc. 

Downsample factors
==================
Downsample : Freq (Hz)   : Timestep period (ms)
1          : 17852.768   : 
2          :  8926.389   :
4          :  4463.192   :  0.224
9          :  1983.641   :  0.504
18         :   991.820   :  1.001
36         :   495.910   :  2.016
71         :   251.447   :  3.977
89         :   200.593   :  4.985
143        :   124.845   :  8.001
179        :    99.736   : 10.026
"""
# Roughly, we are aiming for the following:
downsample_factors = [9, 18, 89, 179]
input_lengths_ms = [1000, 1600]
output_lenghts_ms = [1, 10, 50, 100]

"""
The following are the model input-output configurations we will use. The
three variables are:

    - downsampling factor
    - input length 
    - output length

Given that each model will be trained for each configuration, it is not 
feasible to have a large number of configurations. So some care has been 
given to pick an initial set that will hopefully be interesting and give us
some insight into the problem. After this, we should be able to narrow in
on a smaller set of configurations. 

There are a few considerations driving the choices for these variables.

Downsample factor
-----------------
We are not sure at what point downsampling causes loss of important
information. Ideally, we would like to work with a low sample rate. Trying with
a few different rates will help us to determine the trade-offs at each
downsampling factor. It's likely that further experiments can work with a
single sampling rate.

Input length
------------
I am told that previous experiments have found that retinal ganglion cells can
depend on the past 1200 ms of input. At least one of the papers I am trying to
benchmark against used much shorter inputs (400ms). I am setting a high input
rate of 1600 to give a decent buffer so that we can support/test an argument
that no more than X ms seem to be used, where X is currently hypothesized to be
around 1200. I am setting a low input length of 1000 ms to test if 1000 is
insufficient. The choice of 1000 is somewhat arbitrary. If 1000 is
insufficient, we can justify working with a shorter 1000 ms input, which is a
win from an engineering point of view. If 1000 is insufficient, then this is
nice evidence to support the hypothesis that >1000 ms are used. If we had
chosen 400 ms, this result would not be interesting, as I think it is widely
accepted that the ganglion cells depend on more than the last 400 ms. So 1000
was chosen to try be a win-win: either an engineering win, or a win from the
point of view of having interesting evidence.

Output length
-------------
Varying the output duration is a way to test how temporally precise a model
is. The output duration represents the duration over which spikes will be
summed to calculate the "spike count" for the output interval. In addition,
varying output duration allows us to test our evaluation metrics. For example,
what is the relationship between accuracy, false positives, false negatives,
and any correlation measures. The first set of experiments are using 
[1, 10, 50, 100] ms output durations. The 1 ms output duration is expected to
be too difficult to model, while the 100 ms output is expected to be easy.
So, these two extremes will act as a sanity check, helping to identify any
strange behaviour. The 10 ms output was chosen as I have seen it in other 
literature, so if will be useful for comparison. After these 3 durations,
I wasn't sure what to choose. Given that it's the first set of experiments,
I'm not expecting amazing models, so periods greater than 10 ms might be 
more useful for comparison than the more difficult shorter periods. I hope
we can get to the point where periods between 5 ms and 10 ms are interesting.
"""
all_configs = tuple(
    Configuration(*c)
    for c in [
        # 0.504 ms bins.
        #   1000.18 ms input.
        #        1.008 ms output
        (9, 1984, 2),
        #       10.82 ms output
        (9, 1984, 20),
        #       21.60 ms output
        (9, 1984, 100),
        #       50.41 ms output
        (9, 1984, 198),
        #       99.82 ms output
        #   1600.09 ms input.
        #        1.008 ms output
        (9, 3174, 2),
        #       10.82 ms output
        (9, 3174, 20),
        #       21.60 ms output
        (9, 3174, 100),
        #       50.41 ms output
        (9, 3174, 198),
        #       99.82 ms output
        # 1.001 ms bins.
        #   1000.18 ms input.
        #        1.008 ms output
        (18, 992, 1),
        #       10.08 ms output
        (18, 992, 10),
        #       50.41 ms output
        (18, 992, 50),
        #      100.82 ms output
        (18, 992, 100),
        #   1599.08 ms input.
        #        1.008 ms output
        (18, 1586, 1),
        #       10.08 ms output
        (18, 1586, 10),
        #       50.41 ms output
        (18, 1586, 50),
        #      100.82 ms output
        (18, 1586, 100),
        # Downsample by 89, giving 4.985 ms bins. At this rate, we can't
        # output 1 ms bins, so there are only 6 configurations for this
        # downsample factor.
        # 4.985 ms bins
        #    997.04 ms input
        #    Alternative is the closer 1.002 ms with 201 bins, but going with
        #    201 bins to try and keep the input/output bins even numbers.
        #       9.970 ms output
        (89, 200, 2),
        #       49.85 ms output
        (89, 200, 10),
        #       99.70 ms output
        (89, 200, 20),
        #   1595.27 ms input
        #   Alternative is the 1600.27 ms, with 321 bins, but going with 320
        #   bins to try and keep the input/output bins even numbers.
        #       9.970 ms output
        (89, 320, 2),
        #       49.85 ms output
        (89, 320, 10),
        #       99.70 ms output
        (89, 320, 20),
        # Downsample by 179, giving 10.026 ms bins. Same as with 89, we can't
        # output 1 ms bins, so there are only 6 configurations for this
        # downsample factor.
        # 10.026 ms bins
        #   1002.65 ms input
        #       10.037 ms output
        (179, 100, 1),
        #       20.053 ms output
        (179, 100, 2),
        #       50.132 ms output
        (179, 100, 5),
        #       10.037 ms output
        #   1604.22 ms input
        (179, 160, 1),
        #       20.053 ms output
        (179, 160, 2),
        #       50.132 ms output
        (179, 160, 5),
    ]
)

# To achieve the above, we can run the following function, get_configurations();
# however, the input and output lengths are inconvenient to work with. So,
# we will specify each configuration manually.

# The precise configurations would have been:
#
#     downsample, input bins, output bins
#     9 ds, 1983.6409 in, 1.9836 out
#     9 ds, 1983.6409 in, 19.8364 out
#     9 ds, 1983.6409 in, 99.1820 out
#     9 ds, 1983.6409 in, 198.3641 out
#     9 ds, 3173.8254 in, 1.9836 out
#     9 ds, 3173.8254 in, 19.8364 out
#     9 ds, 3173.8254 in, 99.1820 out
#     9 ds, 3173.8254 in, 198.3641 out
#     18 ds, 991.8204 in, 9.9182 out
#     18 ds, 991.8204 in, 49.5910 out
#     18 ds, 991.8204 in, 99.1820 out
#     18 ds, 1586.9127 in, 9.9182 out
#     18 ds, 1586.9127 in, 49.5910 out
#     18 ds, 1586.9127 in, 99.1820 out
#     89 ds, 200.5929 in, 2.0059 out
#     89 ds, 200.5929 in, 10.0296 out
#     89 ds, 200.5929 in, 20.0593 out
#     89 ds, 320.9486 in, 2.0059 out
#     89 ds, 320.9486 in, 10.0296 out
#     89 ds, 320.9486 in, 20.0593 out
#     etc


def ms_to_num_bins(time_ms, downsample_factor):
    res = time_ms * (mea.ELECTRODE_FREQ / 1000) / downsample_factor
    return res


def get_configurations():
    res = []
    for downsample in downsample_factors:
        for in_len in input_lengths_ms:
            in_bins = ms_to_num_bins(in_len, downsample)
            for out_len in output_lenghts_ms:
                out_bins = ms_to_num_bins(out_len, downsample)
                if in_bins < 1 or out_bins < 1:
                    # Not enough resolution at this downsample factor.
                    continue
                in_bins_int = round(in_bins)
                out_bins_int = round(out_bins)
                res.append(Configuration(downsample, in_bins_int, out_bins_int))
    return res


class LinearNonLinearTrainable(retinapy.train.Trainable):
    def __init__(self, train_ds, val_ds, test_ds, model, model_label):
        super(LinearNonLinearTrainable, self).__init__(
            train_ds, val_ds, test_ds, model, model_label
        )
        self.loss_fn = torch.nn.PoissonNLLLoss(log_input=False)

    def forward(self, sample):
        X, y = sample
        X = X.float().cuda()
        y = y.float().cuda()
        model_output = self.model(X)
        loss = self.loss_fn(model_output, target=y)
        return model_output, loss

    def evaluate(self, val_dl):
        predictions = []
        targets = []
        # Note: the loss per batch is averaged, so we are averaging this again
        # as we loop through each batch.
        loss_meter = retinapy._logging.Meter("loss")
        for (X, y) in val_dl:
            X = X.float().cuda()
            y = y.float().cuda()
            model_output = self.model(X)
            loss_meter.update(
                self.loss_fn(model_output, target=y).item(), y.shape[0]
            )
            predictions.append(model_output.cpu())
            targets.append(y.cpu())
        # Don't forget to check if the model output is log(the Poisson λ parameter)
        # or not log!
        predictions = torch.round(torch.cat(predictions))
        targets = torch.cat(targets)
        acc = (predictions == targets).float().mean().item()
        pearson_corr = scipy.stats.pearsonr(predictions, targets)[0]
        results = {
            "metrics": [
                retinapy._logging.Metric(
                    "loss", loss_meter.avg, increasing=False
                ),
                retinapy._logging.Metric("accuracy", acc),
                retinapy._logging.Metric("pearson_corr", pearson_corr),
            ]
        }
        return results


class DistFieldTrainable_(retinapy.train.Trainable):
    """
    A base trainable for the distance field models.

    There are quite a few things that are common to all such models, such as:
        - the conversion from distance field to model output (and vice versa)
        - the evaluation proceedure
        - the requirement to wrap properties like sample rate that would
          otherwise not be contained in one obvious place.
    """

    eval_lengths_ms: Iterable[int]
    max_eval_count: int

    """
    min_dist has duel purpose:
        a) it represents the fact that a bin that contains a spike means that
           when viewing the bin as a sample point, the actual spike could have
           occurred anywhere ± 1/2 the sample period. Thus, for a bin that
           contains a spike, zero distance is not appropriate—it should be
           the expected value of the distance, which if a uniform distribution
           is assumed, then the distance is 1/2.
        b) it caps the error for a prediction for a bin that contains a spike.
           When using the log distance as the target, a distance of zero will 
           give -∞ as a target. So we should choose some minimum distance. 
           Enter 0.5 as a reasonable minimum given the justification from
           the previous point.

    TODO: I think that due to the explanation in point a) above, the value
    of 1/2 should be set by the dataset, and not here. The reason for delaying
    this now is that quite a bit of changes would need to be made (tests etc.)
    """
    MIN_DIST: float = 0.5
    DEFAULT_EVAL_LENGTHS_MS = [10, 50, 100]
    DEFAULT_MAX_EVAL_COUNT = int(5e5)

    def __init__(
        self,
        train_ds,
        val_ds,
        test_ds,
        model,
        model_label,
        eval_lengths=None,
    ):
        super().__init__(train_ds, val_ds, test_ds, model, model_label)
        sample_rates_equal = (
            train_ds.sample_rate == test_ds.sample_rate == val_ds.sample_rate
        )
        if not sample_rates_equal:
            raise ValueError(
                "Train, validation and test datasets do not have "
                f"the same sample rate. Got ({train_ds.sample_rate:.5f}, "
                f"{val_ds.sample_rate:.5f}, {test_ds.sample_rate:.5f})"
            )
        if eval_lengths is None:
            self.eval_lengths_ms = self.DEFAULT_EVAL_LENGTHS_MS
        else:
            self.eval_lengths_ms = eval_lengths
        self.max_eval_count = self.DEFAULT_MAX_EVAL_COUNT
        # Insure the model output has a mean not too far from 0.
        MAX_MODEL_OUTPUT = 3.4
        self.output_offset = (
            # log(max_distance) - OFFSET = MAX_MODEL_OUTPUT
            math.log(self.ms_to_bins(DIST_CLAMP_MS))
            - MAX_MODEL_OUTPUT
        )
        self.num_plots = 16

    @property
    def in_device(self):
        res = next(self.model.parameters()).device
        return res

    def ms_to_bins(self, ms: float) -> int:
        num_bins = max(1, round(ms * (self.sample_rate / 1000)))
        return num_bins

    @property
    def max_bin_dist(self):
        return self.ms_to_bins(DIST_CLAMP_MS)

    @property
    def sample_period_ms(self):
        return 1000 / self.sample_rate

    @property
    def sample_rate(self):
        sample_rates_equal = (
            self.train_ds.sample_rate
            == self.test_ds.sample_rate
            == self.val_ds.sample_rate
        )
        assert sample_rates_equal
        # We can use any of the datasets.
        return self.train_ds.sample_rate

    def distfield_to_nn_output(self, distfield):
        return torch.log((distfield + self.MIN_DIST)) - self.output_offset

    def nn_output_to_distfield(self, nn_output):
        return torch.exp(nn_output + self.output_offset) - self.MIN_DIST

    def quick_infer(self, dist, num_bins):
        """Quickly infer the number of spikes in the eval region.

        An approximate inference used for evaluation.

        Returns:
            the number of spikes in the region.
        """
        threshold = -0.4
        res = (dist[:, 0:num_bins] < threshold).sum(dim=1)
        return res

    def evaluate(self, val_dl):
        predictions = defaultdict(list)
        targets = defaultdict(list)
        loss_meter = retinapy._logging.Meter("loss")
        plotly_figs = []
        for i, sample in enumerate(val_dl):
            # Don't run out of memory, or take too long.
            num_so_far = val_dl.batch_size * i
            if num_so_far > self.max_eval_count:
                break
            target_spikes = sample["target_spikes"].float().cuda()
            model_output, loss = self.forward(sample)
            loss_meter.update(loss.item())
            # Count accuracies
            for eval_len in self.eval_lengths_ms:
                eval_bins = self.ms_to_bins(eval_len)
                pred = self.quick_infer(model_output, num_bins=eval_bins)
                y = torch.sum(target_spikes[:, 0:eval_bins], dim=1)
                predictions[eval_len].append(pred)
                targets[eval_len].append(y)
            # Plot some example input-outputs. Only for snippets with spikes,
            # as otherwise the plots are not very interesting.
            if len(plotly_figs) < self.num_plots and torch.sum(target_spikes):
                # Plot the first batch element.
                idx = 0
                plotly_figs.append(
                    self.input_output_fig(sample, model_output, idx)
                )

        metrics = [
            retinapy._logging.Metric("loss", loss_meter.avg, increasing=False)
        ]
        for eval_len in self.eval_lengths_ms:
            p = torch.cat(predictions[eval_len])
            t = torch.cat(targets[eval_len])
            acc = (p == t).float().mean().item()
            pearson_corr = scipy.stats.pearsonr(
                p.cpu().numpy(), t.cpu().numpy()
            )[0]
            metrics.append(
                retinapy._logging.Metric(f"accuracy-{eval_len}_ms", acc)
            )
            metrics.append(
                retinapy._logging.Metric(
                    f"pearson_corr-{eval_len}_ms", pearson_corr
                )
            )
        results = {
            "metrics": metrics,
            "input-output-figs": retinapy._logging.PlotlyFigureList(
                plotly_figs
            ),
        }
        return results

    def input_output_fig(self, sample, model_out, idx):
        target_dist = self.distfield_to_nn_output(sample["dist"][idx])

        def cluster_label(sample):
            """Get the cluster label for a sample.

            Used to allow both singel and multi-cluster datasets to use a
            shared evaluation function.
            """
            if "rec_id" in sample:
                res = (
                    f"rec idx: {sample['rec_id'][idx]},"
                    f"cluster idx: {sample['cluster_id'][idx]}"
                )
            else:
                res = f"cluster idx: {sample['cluster_id'][idx]}"
            return res

        fig = retinapy.vis.distfield_model_in_out(
            # Stimulus takes up all channels except the last.
            stimulus=sample["snippet"][idx][0:4].cpu().numpy(),
            in_spikes=sample["snippet"][idx][-1].cpu().numpy(),
            target_dist=target_dist.cpu().numpy(),
            model_out=model_out[idx].cpu().numpy(),
            start_ms=0,
            bin_duration_ms=self.sample_period_ms,
            cluster_label=cluster_label(sample),
        )
        # The legend takes up a lot of space, so disable it.
        fig.update_layout(showlegend=False)
        return fig


class DistFieldVAETrainable(DistFieldTrainable_):
    def __init__(
        self,
        train_ds,
        val_ds,
        test_ds,
        model,
        model_label,
        eval_lengths,
        vae_beta,
    ):
        """Trainable for a multi-cluster distance field model.

        This is a minimal extension of the DistFieldTrainable.
        """
        super().__init__(
            train_ds, val_ds, test_ds, model, model_label, eval_lengths
        )
        self.dist_loss_fn = retinapy.models.dist_loss
        # Network output should ideally have mean,sd = (0, 1). Network output
        # 20*exp([-3, 3])  = [1.0, 402], which is a pretty good range, with
        # 20 being the mid point. Is this too low?
        self.dist_norm = 20
        self.vae_beta = vae_beta
        # Network output should ideally have mean,sd = (0, 1). Network output
        # 20*exp([-3, 3])  = [1.0, 402], which is a pretty good range, with
        # 20 being the mid point. Is this too low?
        self.max_eval_count = int(2e5)

    def loss(self, m_dist, z_mu, z_lorvar, target):
        batch_size = m_dist.shape[0]
        # Scale to get roughly in the ballpark of 1.
        dist_loss = self.dist_loss_fn(m_dist, target)
        kl_loss = -0.5 * torch.sum(1 + z_lorvar - z_mu.pow(2) - z_lorvar.exp())
        dist_loss = dist_loss / batch_size
        # β = 1/1000
        β = self.vae_beta
        kl_loss = β * kl_loss / batch_size
        total = dist_loss + kl_loss
        return total, dist_loss, kl_loss

    def all_clusters(self):
        """
        Return a Tensor where every row is (rec_idx, cluster_idx, cluster_id).
        """
        res = []
        # This depends on the dataset being a concat dataset.
        for idx, rec_ds in enumerate(self.train_ds.datasets):
            recording = rec_ds.recording
            c_idxs = torch.arange(len(recording.cluster_ids))
            rec_idx_cidx_cid = torch.stack(
                [
                    torch.full_like(c_idxs, idx),
                    c_idxs,
                    torch.tensor(recording.cluster_ids, device=c_idxs.device),
                ],
                dim=-1,
            )
            res.append(rec_idx_cidx_cid)
        res = torch.cat(res)
        return res

    def model_summary(self, sample):
        masked_snippet = (
            sample["snippet"][:, 0 : mea.NUM_STIMULUS_LEDS + 1]
            .float()
            .to(self.in_device)
        )
        rec_id = sample["rec_id"].to(self.in_device)
        cluster_id = sample["cluster_id"].to(self.in_device)
        res = torchinfo.summary(
            self.model,
            input_data=[masked_snippet, rec_id, cluster_id],
            col_names=["input_size", "output_size", "mult_adds", "num_params"],
            device=self.in_device,
            depth=4,
        )
        return res

    def encode(
        self, rec_idxs: torch.LongTensor, cluster_idxs: torch.LongTensor
    ):
        """
        Return the latent representation of the given recording-cluster pairs.

        Args:
            rec_idxs: a batched tensor of recording indexes.
            cluster_idxs: a batch tensor of cluster indexes.
        """
        # We wrap the to-from device calls here, just for a trial.
        # Maybe we should make a separate non-sampling function, but it's fine
        # for now.
        in_device = rec_idxs.device
        rec_idxs = rec_idxs.to(self.in_device)
        cluster_idxs = cluster_idxs.to(self.in_device)
        _, z, _ = self.model.encode_vae(rec_idxs, cluster_idxs)
        return z.to(in_device)

    def all_encodings(self):
        all_clusters = self.all_clusters()
        zs = self.encode(all_clusters[:, 0], all_clusters[:, 1])
        res = torch.cat([all_clusters, zs], dim=-1)
        return res

    def _forward(self, sample):
        masked_snippet = sample["snippet"].float().to(self.in_device)
        dist = sample["dist"].float().to(self.in_device)
        rec_id = sample["rec_id"].int().to(self.in_device)
        cluster_id = sample["cluster_id"].int().to(self.in_device)
        m_dist, z_mu, z_logvar = self.model(masked_snippet, rec_id, cluster_id)
        # Dist model
        y = self.distfield_to_nn_output(dist)
        loss, dist_loss, kl_loss = self.loss(m_dist, z_mu, z_logvar, target=y)
        return m_dist, loss, dist_loss, kl_loss

    def forward(self, sample):
        m_dist, loss, _, _ = self._forward(sample)
        return m_dist, loss

    def evaluate(self, val_dl):
        predictions = defaultdict(list)
        targets = defaultdict(list)
        loss_meter = retinapy._logging.Meter("loss")
        kl_loss_meter = retinapy._logging.Meter("kl-loss")
        input_output_figs = []
        for i, sample in enumerate(val_dl):
            # Don't run out of memory, or take too long.
            num_so_far = val_dl.batch_size * i
            if num_so_far > self.max_eval_count:
                break
            target_spikes = sample["target_spikes"].float().cuda()
            model_output, loss, _, kl_loss = self._forward(sample)
            loss_meter.update(loss.item())
            kl_loss_meter.update(kl_loss.item())
            # Count accuracies
            for eval_len in self.eval_lengths_ms:
                eval_bins = self.ms_to_bins(eval_len)
                pred = self.quick_infer(model_output, num_bins=eval_bins)
                y = torch.sum(target_spikes[:, 0:eval_bins], dim=1)
                predictions[eval_len].append(pred)
                targets[eval_len].append(y)
            # Plot some example input-outputs
            if len(input_output_figs) < self.num_plots:
                # Plot the first batch element.
                idx = 0
                # Don't bother if there is no spike.
                contains_spike = torch.sum(target_spikes[0]) > 0
                if contains_spike:
                    input_output_figs.append(
                        self.input_output_fig(sample, model_output, idx)
                    )

        metrics = [
            retinapy._logging.Metric("loss", loss_meter.avg, increasing=False),
            retinapy._logging.Metric(
                "kl-loss", kl_loss_meter.avg, increasing=False
            ),
        ]
        for eval_len in self.eval_lengths_ms:
            p = torch.cat(predictions[eval_len])
            t = torch.cat(targets[eval_len])
            acc = (p == t).float().mean().item()
            pearson_corr = scipy.stats.pearsonr(
                p.cpu().numpy(), t.cpu().numpy()
            )[0]
            metrics.append(
                retinapy._logging.Metric(f"accuracy-{eval_len}_ms", acc)
            )
            metrics.append(
                retinapy._logging.Metric(
                    f"pearson_corr-{eval_len}_ms", pearson_corr
                )
            )
        results = {
            "metrics": metrics,
            "input-output-figs": retinapy._logging.PlotlyFigureList(
                input_output_figs
            ),
        }
        # Add the latent space visualization.
        encodings = self.all_encodings()
        rec_ids, _, cluster_ids, *zs = list(encodings.T.cpu().numpy())
        if self.model.z_dim == 2:
            latent_fig = retinapy.vis.latent2d_fig(
                rec_ids, cluster_ids, zs[0], zs[1]
            )
            results["latent-fig"] = retinapy._logging.PlotlyFigureList(
                [latent_fig]
            )
        # And try out Tensorboard's embedding feature.
        results["z-embeddings"] = retinapy._logging.Embeddings(
            embeddings=encodings[:, 2:].cpu().numpy(),
            labels=[
                f"{int(r_id)}-{int(c_id)}"
                for r_id, c_id in zip(rec_ids, cluster_ids)
            ],
        )
        return results


class CnnDistFieldTrainable(DistFieldTrainable_):
    def __init__(
        self, train_ds, val_ds, test_ds, model, model_label, eval_lengths
    ):
        """Trainable for a distance field model.

        Notable is the eval_len parameter (not present in Trainable), which
        is number of bins to consider when claculating accuracy. This is
        needed as the you typically want to guess the number of spikes in a
        region by using a distance field that is bigger than and contains the
        region. So there is not a 1-1 between distance field length and the
        number of bins over which we are counting spikes.
        """
        super().__init__(
            train_ds, val_ds, test_ds, model, model_label, eval_lengths
        )

    def loss(self, m_dist, target):
        batch_size = m_dist.shape[0]
        batch_sum = retinapy.models.dist_loss(m_dist, target)
        batch_ave = batch_sum / batch_size
        return batch_ave

    def forward(self, sample):
        masked_snippet = sample["snippet"].float().cuda()
        dist = sample["dist"].float().cuda()
        model_output = self.model(masked_snippet)
        # Dist model
        y = self.distfield_to_nn_output(dist)
        loss = self.loss(model_output, target=y)
        return model_output, loss

    def model_summary(self, sample):
        masked_snippet = sample["snippet"].float().cuda()
        res = torchinfo.summary(
            self.model,
            input_data=masked_snippet,
            col_names=["input_size", "output_size", "mult_adds", "num_params"],
            device=self.in_device,
            depth=4,
        )
        return res


def create_multi_cluster_df_datasets(
    recordings: Iterable[mea.CompressedSpikeRecording],
    input_len: int,
    output_len: int,
    downsample: int,
    stride: int,
    num_workers: int,
):
    train_ds = []
    val_ds = []
    test_ds = []
    # Make a queue to save memory by deleting while iterating.
    dc_recs_queue = mea.decompress_recordings(
        recordings, downsample=downsample, num_workers=num_workers
    )
    while dc_recs_queue:
        rec = dc_recs_queue.popleft()
        train_val_test_splits = mea.mirror_split(rec, split_ratio=SPLIT_RATIO)
        snippet_len = input_len + output_len
        train_val_test_datasets = [
            retinapy.dataset.DistFieldDataset(
                r,
                snippet_len=snippet_len,
                mask_begin=input_len,
                mask_end=snippet_len,
                pad=round(ms_to_num_bins(LOSS_CALC_PAD_MS, downsample)),
                dist_clamp=round(ms_to_num_bins(DIST_CLAMP_MS, downsample)),
                stride=stride,
                enable_augmentation=use_augmentation,
                allow_cheating=False,
            )
            for (r, use_augmentation) in zip(
                # train_val_test_splits, [True, False, False]
                # For the moment, while the validation just takes a small
                # portion of the validation set.
                train_val_test_splits,
                [True, True, False],
            )
        ]
        train_ds.append(train_val_test_datasets[0])
        val_ds.append(train_val_test_datasets[1])
        test_ds.append(train_val_test_datasets[2])
    concat_train_ds = retinapy.dataset.ConcatDistFieldDataset(train_ds)
    concat_val_ds = retinapy.dataset.ConcatDistFieldDataset(val_ds)
    concat_test_ds = retinapy.dataset.ConcatDistFieldDataset(test_ds)
    res = (concat_train_ds, concat_val_ds, concat_test_ds)
    return res


def create_distfield_datasets(
    recording: mea.CompressedSpikeRecording,
    input_len: int,
    output_len: int,
    downsample: int,
):
    rec = mea.decompress_recording(recording, downsample=downsample)
    train_val_test_splits = mea.mirror_split(rec, split_ratio=SPLIT_RATIO)
    snippet_len = input_len + output_len
    train_val_test_datasets = [
        retinapy.dataset.DistFieldDataset(
            r,
            snippet_len=snippet_len,
            mask_begin=input_len,
            mask_end=snippet_len,
            pad=round(ms_to_num_bins(LOSS_CALC_PAD_MS, downsample)),
            dist_clamp=round(ms_to_num_bins(DIST_CLAMP_MS, downsample)),
            enable_augmentation=use_augmentation,
            allow_cheating=False,
        )
        for (r, use_augmentation) in zip(
            train_val_test_splits, [True, False, False]
        )
    ]
    return train_val_test_datasets


def create_count_datasets(
    recording: mea.CompressedSpikeRecording,
    input_len: int,
    output_len: int,
    downsample: int,
):
    """
    Creates the spike count datasets for the given recording data.

    Three datasets are returned: train, validation and test.

    These datasets take the form:
        X,y = (stimulus-spike history, num spikes)

    The length of the input history, the output binning duration and the
    downsample rate can be configured.
    """
    rec = mea.decompress_recording(recording, downsample=downsample)
    train_val_test_splits = mea.mirror_split(rec, split_ratio=SPLIT_RATIO)
    # train_val_test_splits = mea.split(rec, split_ratio=SPLIT_RATIO)
    train_val_test_datasets = [
        retinapy.dataset.SpikeCountDataset(
            r,
            input_len=input_len,
            output_len=output_len,
        )
        for r in train_val_test_splits
    ]
    return train_val_test_datasets


class TrainableGroup:
    def trainable_label(self, config):
        raise NotImplementedError

    def create_trainable(
        self, recordings: Iterable[mea.CompressedSpikeRecording], config, opt
    ):
        raise NotImplementedError


class MultiClusterDistFieldTGroup(TrainableGroup):
    @staticmethod
    def trainable_label(config):
        return (
            f"MultiClusterDistField-{config.downsample}"
            f"ds_{config.input_len}in_{config.output_len}out"
        )

    @staticmethod
    def model_output_len(config) -> int:
        """Determine the model output length for the required eval length.

        Currently, it's double the eval length. Large model length provides
        more information to the model while training, and provides the option
        """
        min_model_output_ms = 50
        min_model_output_bins = ms_to_num_bins(
            min_model_output_ms, config.downsample
        )
        eval_len_bins = ms_to_num_bins(config.output_len, config.downsample)
        model_bins = round(max(min_model_output_bins, eval_len_bins * 2))
        return model_bins

    @classmethod
    def create_trainable(cls, recordings, config, opt):
        output_len = cls.model_output_len(config)
        train_ds, val_ds, test_ds = create_multi_cluster_df_datasets(
            recordings,
            config.input_len,
            output_len,
            config.downsample,
            stride=opt.stride,
            num_workers=opt.num_workers,
        )
        # There is separation between the target inference duration, say 10ms,
        # and the output length of the model, say 20ms. The model output should
        # be at least as large as the target inference duration, and it will
        # probably benefit in being larger.
        max_num_clusters = max([len(r.cluster_ids) for r in recordings])
        model = retinapy.models.CatMultiClusterModel(
            config.input_len + output_len,
            output_len,
            cls.num_downsample_layers(config.input_len, config.output_len),
            len(recordings),
            max_num_clusters,
            z_dim=opt.zdim,
        )
        res = DistFieldVAETrainable(
            train_ds,
            val_ds,
            test_ds,
            model,
            DistFieldCnnTGroup.trainable_label(config),
            eval_lengths=[10, 20, 50],
            vae_beta=opt.vae_beta,
        )
        return res

    @staticmethod
    def num_downsample_layers(in_len, out_len):
        res_f = math.log(in_len / out_len, 2)
        num_downsample = int(res_f)
        if res_f - num_downsample > 0.4:
            logging.warning(
                "Model input/output lengths are not well matched. "
                f"Downsample desired: ({res_f}), downsample being used: "
                f"({num_downsample})."
            )
        return num_downsample


class TransformerTGroup(TrainableGroup):
    @staticmethod
    def trainable_label(config):
        return f"Transformer-{config.downsample}" f"ds_{config.input_len}in"

    @staticmethod
    def create_trainable(
        recordings, config, opt, num_recs=None, max_num_clusters=None
    ):
        # There is separation between the target inference duration, say 10ms,
        # and the output length of the model, say 20ms. The model output should
        # be at least as large as the target inference duration, and it will
        # probably benefit in being larger.
        model_out_len = {1984: 200, 992: 150, 3174: 400, 1586: 200}[
            config.input_len
        ]
        stim_ds = {1984: 6, 992: 5, 3174: 7, 1586: 6}[config.input_len]
        spike_patch_len = {1984: 16, 992: 8, 3174: 32, 1586: 16}[
            config.input_len
        ]
        train_ds, val_ds, test_ds = create_multi_cluster_df_datasets(
            recordings,
            config.input_len,
            model_out_len,
            config.downsample,
            stride=opt.stride,
            num_workers=opt.num_workers,
        )
        num_recs = num_recs if num_recs else len(recordings)
        calc_max_num_clusters = max([len(r.cluster_ids) for r in recordings])
        max_num_clusters = (
            max_num_clusters if max_num_clusters else calc_max_num_clusters
        )
        model = retinapy.models.TransformerModel(
            config.input_len,
            model_out_len,
            stim_downsample=stim_ds,
            num_recordings=num_recs,
            num_clusters=max_num_clusters,
            z_dim=opt.zdim,
            num_heads=opt.num_heads,
            head_dim=opt.head_dim,
            num_tlayers=opt.num_tlayers,
            spike_patch_len=spike_patch_len,
        )
        res = DistFieldVAETrainable(
            train_ds,
            val_ds,
            test_ds,
            model,
            DistFieldCnnTGroup.trainable_label(config),
            eval_lengths=[10, 20, 50],
            vae_beta=opt.vae_beta,
        )
        return res


class ClusteringTGroup(TrainableGroup):
    @staticmethod
    def trainable_label(config):
        return (
            f"ClusterTransformer-{config.downsample}" f"ds_{config.input_len}in"
        )

    @staticmethod
    def create_trainable(
        recordings, config, opt, num_recs=None, max_num_clusters=None
    ):
        # There is separation between the target inference duration, say 10ms,
        # and the output length of the model, say 20ms. The model output should
        # be at least as large as the target inference duration, and it will
        # probably benefit in being larger.
        model_out_len = {1984: 200, 992: 150, 3174: 400, 1586: 200}[
            config.input_len
        ]
        stim_ds = {1984: 6, 992: 5, 3174: 7, 1586: 6}[config.input_len]
        train_ds, val_ds, test_ds = create_multi_cluster_df_datasets(
            recordings,
            config.input_len,
            model_out_len,
            config.downsample,
            stride=opt.stride,
            num_workers=opt.num_workers,
        )
        num_recs = num_recs if num_recs else len(recordings)
        calc_max_num_clusters = max([len(r.cluster_ids) for r in recordings])
        max_num_clusters = (
            max_num_clusters if max_num_clusters else calc_max_num_clusters
        )
        model = retinapy.models.ClusteringTransformer(
            config.input_len,
            model_out_len,
            stim_downsample=stim_ds,
            num_recordings=num_recs,
            num_clusters=max_num_clusters,
            z_dim=opt.zdim,
            num_heads=opt.num_heads,
            head_dim=opt.head_dim,
            num_tlayers=opt.num_tlayers,
        )
        res = DistFieldVAETrainable(
            train_ds,
            val_ds,
            test_ds,
            model,
            DistFieldCnnTGroup.trainable_label(config),
            eval_lengths=[10, 20, 50],
            vae_beta=opt.vae_beta,
        )
        return res


class DistFieldCnnTGroup(TrainableGroup):
    @staticmethod
    def trainable_label(config):
        return f"DistFieldCnn-{config.downsample}" f"ds_{config.input_len}in"

    @staticmethod
    def create_trainable(recordings, config, opt):
        if len(recordings) != 1 or len(recordings[0].cluster_ids) != 1:
            raise ValueError(
                "DistFieldCnn model only supports a single cluster."
            )
        rec = recordings[0]
        assert len(rec.cluster_ids) == 1
        # There is separation between the target inference duration, say 10ms,
        # and the output length of the model, say 20ms. The model output should
        # be at least as large as the target inference duration, and it will
        # probably benefit in being larger.
        output_lens = {1984: 200, 992: 100, 3174: 400, 1586: 200}
        model_out_len = output_lens[config.input_len]
        train_ds, val_ds, test_ds = create_distfield_datasets(
            rec, config.input_len, model_out_len, config.downsample
        )
        model = retinapy.models.DistanceFieldCnnModel(
            config.input_len + model_out_len,
            model_out_len,
        )
        res = CnnDistFieldTrainable(
            train_ds,
            val_ds,
            test_ds,
            model,
            DistFieldCnnTGroup.trainable_label(config),
            eval_lengths=[5, 10, 20, 50, 100],
        )
        return res


class LinearNonLinearTGroup(TrainableGroup):
    @staticmethod
    def trainable_label(config):
        return (
            f"LinearNonLinear-{config.downsample}"
            f"ds_{config.input_len}in_{config.output_len}out"
        )

    @staticmethod
    def create_trainable(recordings, config, opt):
        rec = recordings[0]
        num_inputs = IN_CHANNELS * config.input_len
        m = retinapy.models.LinearNonlinear(in_n=num_inputs, out_n=1)
        train_ds, val_ds, test_ds = create_count_datasets(
            rec, config.input_len, config.output_len, config.downsample
        )
        label = LinearNonLinearTGroup.trainable_label(config)
        return LinearNonLinearTrainable(train_ds, val_ds, test_ds, m, label)


def _train(out_dir, opt):
    print("Models & Configurations")
    print("=======================")
    # Product of models and configs
    trainable_groups = {
        "LinearNonLinear": LinearNonLinearTGroup,
        "DistFieldCnn": DistFieldCnnTGroup,
        "MultiClusterDistField": MultiClusterDistFieldTGroup,
        "Transformer": TransformerTGroup,
        "ClusterTransformer": ClusteringTGroup,
    }

    def run_id(model_str, config):
        return f"{model_str}-{str(config)}"

    def _match(run_id, match_str):
        return match_str in run_id

    do_trainable = dict()
    for c in all_configs:
        for _, tg in trainable_groups.items():
            t_label = tg.trainable_label(c)
            do_trainable[t_label] = _match(t_label, opt.k)

    logging.info(f"Model-configs filter: {opt.k}")
    logging.info(
        "\n".join(
            [
                t_label if do_train else t_label.ljust(40) + " (skip)"
                for t_label, do_train in do_trainable.items()
            ]
        )
    )
    total_trainables = sum(do_trainable.values())
    logging.info(f"Total: {total_trainables} models to be trained.")

    # Load the data.
    # Filter recordings, if requested.
    if opt.recording_names is not None and len(opt.recording_names) == 1:
        # If only one recording, then cluster-ids can be specified.
        if opt.cluster_ids is not None:
            include_cluster_ids = set(opt.cluster_ids)
        else:
            include_cluster_ids = None
        recordings = [
            mea.single_3brain_recording(
                opt.recording_names[0],
                opt.data_dir,
                include_clusters=include_cluster_ids,
            )
        ]
    else:
        recordings = mea.load_3brain_recordings(opt.data_dir)
        ## Filter the recording with non-standard sample rate
        skip_rec_names = {"Chicken_21_08_21_Phase_00"}
        recordings = [r for r in recordings if r.name not in skip_rec_names]

    done_trainables = set()
    for c in all_configs:
        for tg in trainable_groups.values():
            t_label = tg.trainable_label(c)
            if t_label in done_trainables:
                continue
            if not do_trainable[t_label]:
                continue
            t = tg.create_trainable(recordings, c, opt)
            if t is None:
                logging.warning(
                    f"Skipping. Model ({t_label}) isn't yet supported."
                )
                continue
            num_done = len(done_trainables)
            logging.info(
                f"Starting model training ({num_done}/{total_trainables}): "
                f"{t_label}"
            )
            sub_dir = out_dir / str(t_label)
            logging.info(f"Output directory: ({sub_dir})")
            retinapy.train.train(
                t,
                num_epochs=opt.epochs,
                batch_size=opt.batch_size,
                lr=opt.lr,
                weight_decay=opt.weight_decay,
                out_dir=sub_dir,
                steps_til_log=opt.steps_til_log,
                steps_til_eval=opt.steps_til_eval,
                initial_checkpoint=opt.initial_checkpoint,
                num_workers=opt.num_workers,
                pin_memory=opt.pin_memory,
            )
            logging.info(f"Finished training model")
            done_trainables.add(t_label)
    logging.info("Finished training all linear non-linear models.")


def main():
    retinapy._logging.setup_logging(logging.INFO)
    opt, opt_text = parse_args()
    labels = opt.labels.split(",") if opt.labels else None
    base_dir = pathlib.Path(opt.output if opt.output else DEFAULT_OUT_BASE_DIR)
    out_dir = retinapy._logging.get_outdir(base_dir, labels)
    print("Output directory:", out_dir)
    retinapy._logging.enable_file_logging(out_dir / LOG_FILENAME)
    # Record the arguments.
    with open(str(out_dir / ARGS_FILENAME), "w") as f:
        f.write(opt_text)
    _train(out_dir, opt)


if __name__ == "__main__":
    main()



import logging
import math
import pickle

import numpy as np
import numpy.ma as ma
import numpy.testing
import pandas as pd

import pytest
import retinapy.mea as mea
import pathlib


DATA_DIR = pathlib.Path("./data/ff_noise_recordings")
FF_NOISE_PATTERN_PATH = DATA_DIR / "stimulus_pattern.h5"
FF_SPIKE_RESPONSE_PATH = DATA_DIR / "spike_response.pickle"
FF_SPIKE_RESPONSE_PATH_ZIP = DATA_DIR / "spike_response.pickle.zip"
FF_RECORDED_NOISE_PATH = DATA_DIR / "recorded_stimulus.pickle"
FF_RECORDED_NOISE_PATH_ZIP = DATA_DIR / "recorded_stimulus.pickle.zip"


def test_load_stimulus_pattern():
    noise = mea.load_stimulus_pattern(FF_NOISE_PATTERN_PATH)
    known_shape = (24000, 4)
    assert noise.shape == known_shape


def test_load_response():
    for path in (FF_SPIKE_RESPONSE_PATH, FF_SPIKE_RESPONSE_PATH_ZIP):
        response = mea.load_response(path)
        known_index_names = ["Cell index", "Stimulus ID", "Recording"]
        assert response.index.names == known_index_names
        known_shape = (4417, 2)
        assert response.shape == known_shape


def test_load_recorded_stimulus():
    for path in (FF_RECORDED_NOISE_PATH, FF_RECORDED_NOISE_PATH_ZIP):
        res = mea.load_recorded_stimulus(path)
        known_index_names = ["Stimulus_index", "Recording"]
        assert res.index.names == known_index_names
        known_shape = (18, 8)
        assert res.shape == known_shape


@pytest.fixture
def stimulus_pattern():
    return mea.load_stimulus_pattern(FF_NOISE_PATTERN_PATH)


@pytest.fixture
def recorded_stimulus():
    return mea.load_recorded_stimulus(FF_RECORDED_NOISE_PATH)


@pytest.fixture
def response_data():
    return mea.load_response(FF_SPIKE_RESPONSE_PATH)


def test_recording_names(response_data):
    known_list = [
        "Chicken_04_08_21_Phase_01",
        "Chicken_04_08_21_Phase_02",
        "Chicken_05_08_21_Phase_00",
        "Chicken_05_08_21_Phase_01",
        "Chicken_06_08_21_2nd_Phase_00",
        "Chicken_06_08_21_Phase_00",
        "Chicken_11_08_21_Phase_00",
        "Chicken_12_08_21_Phase_00",
        "Chicken_12_08_21_Phase_02",
        "Chicken_13_08_21_Phase_00",
        "Chicken_13_08_21_Phase_01",
        "Chicken_14_08_21_Phase_00",
        "Chicken_17_08_21_Phase_00",
        "Chicken_19_08_21_Phase_00",
        "Chicken_19_08_21_Phase_01",
        "Chicken_20_08_21_Phase_00",
        "Chicken_21_08_21_Phase_00",
    ]
    rec_list = mea.recording_names(response_data)
    assert rec_list == known_list


def test_cluster_ids(response_data):
    # fmt: off
    known_list = [12, 13, 14, 15, 17, 25, 28, 29, 34, 44, 45, 50, 60, 61, 80,
                  82, 99, 114, 119, 149, 217, 224, 287, 317, 421, 553, 591]
    # fmt: on
    recording_name = "Chicken_21_08_21_Phase_00"
    cluster_ids = mea._cluster_ids(response_data, recording_name)
    assert cluster_ids == known_list


def test_load_3brain_recordings():
    # Test
    res = mea.load_3brain_recordings(
        DATA_DIR,
        include=[
            "Chicken_04_08_21_Phase_01",
            "Chicken_04_08_21_Phase_02",
        ],
    )
    assert len(res) == 2


def test_split(dc_rec12):
    """Tests splitting a recording into multiple parts.

    Tests that:
        1. A split works.
        2. zero-value ratio causes an error.
    """
    # Test 1
    # Setup
    splits = (3, 1, 1)
    expected_len = 892863
    assert len(dc_rec12) == expected_len
    expected_split_lens = [535716 + 3, 178572, 178572]
    expected_split_lens_reversed = [178572 + 3, 178572, 535716]
    assert len(dc_rec12) == sum(expected_split_lens)
    # Test
    res = mea.split(dc_rec12, splits)
    assert len(res) == 3, "There should be 3 splits."
    assert [
        len(s) for s in res
    ] == expected_split_lens, "Splits should be the correct length."
    # Do it again but reversed.
    res = mea.split(dc_rec12, splits[::-1])
    assert len(splits) == 3, "There should be 3 splits."
    assert [
        len(s) for s in res
    ] == expected_split_lens_reversed, "Splits should be the correct length."

    # Test 2
    with pytest.raises(ValueError):
        mea.split(dc_rec12, (0, 1, 1))


def test_mirror_split(dc_rec12):
    """Tests splitting a recording into multiple parts (the mirrored version).

    Tests that:
        1. A (3, 1, 1) and (1, 1, 3) split works.
            - no errors
            - number of output datasets is correct
            - dataset sizes are as expected
        2. zero-value ratio causes an error.
        3. requesting less than two splits causes an error.
    """
    # Test 1
    # Setup
    splits = (3, 1, 1)
    expected_len = 892863
    assert len(dc_rec12) == expected_len
    # Note how the remainders fall in different places compared to the
    # non-mirrored split. This is due to the mirrored split calling split
    # twice, under the hood.
    expected_split_lens = [535716 + 1, 178572, 178572 + 2]
    expected_split_lens_reversed = [178572 + 1, 178572, 535716 + 2]
    assert len(dc_rec12) == sum(expected_split_lens)
    # Test
    res = mea.mirror_split(dc_rec12, splits)
    assert len(res) == 3, "There should be 3 splits."
    assert [
        len(s) for s in res
    ] == expected_split_lens, "Splits should be the correct length."
    # Do it again but reversed.
    res = mea.mirror_split(dc_rec12, splits[::-1])
    assert len(splits) == 3, "There should be 3 splits."
    assert [
        len(s) for s in res
    ] == expected_split_lens_reversed, "Splits should be the correct length."

    # Test 2
    with pytest.raises(ValueError):
        mea.split(dc_rec12, (0, 1, 1))

    # Test 3
    with pytest.raises(ValueError):
        mea.split(dc_rec12, (1,))
    with pytest.raises(ValueError):
        mea.split(dc_rec12, tuple())


def test_mirror_split2(dc_rec12):
    """
    Invariant test for mirror_split.

    The other test inspects a specific case; here we test various splits and
    check for expected invariants.

    Tests that:
        - split lengths sum up to the length of the original.
        - splits have the same number of clusters.
        - the values (stimulus and spikes) in the first half of the first
            split match the beginning of the original.
    """
    # Setup
    seed = 123
    rng = np.random.default_rng(seed)
    num_trials = 20
    orig_len = len(dc_rec12)
    num_splits = rng.integers(low=2, high=10, size=num_trials)
    ratios = [
        rng.integers(low=1, high=20, size=num_splits[i])
        for i in range(num_trials)
    ]

    def check_sizes(ratio, splits):
        # The combined splits have the same number of timesteps as the original.
        total_len = sum([len(s) for s in splits])
        assert total_len == orig_len
        # The cluster count is also the same.
        expected_num_clusters = len(dc_rec12.cluster_ids)
        for split in splits:
            assert len(split.cluster_ids) == expected_num_clusters

    def check_split1_values(ratio, splits):
        split1_len = splits[0].stimulus.shape[0]
        test_len = split1_len // 2
        orig_stim = dc_rec12.stimulus[:test_len]
        orig_spikes = dc_rec12.spikes[:test_len]
        np.testing.assert_array_equal(orig_stim, splits[0].stimulus[:test_len])
        np.testing.assert_array_equal(orig_spikes, splits[0].spikes[:test_len])

    # Test
    for i in range(num_trials):
        splits = mea.mirror_split(dc_rec12, ratios[i].tolist())
        check_sizes(ratios[i], splits)
        check_split1_values(ratios[i], splits)


def test_decompress_recording(rec12):
    """
    Test decompressing a recording.

    Tests that:
        1. Simple decompression works (basic checks).
        2. Multiple spikes in the same bucket is allowed.
    """
    # Test 1
    downsample = 18
    orig_freq = rec12.sensor_sample_rate
    expected_sample_rate = orig_freq / downsample
    res = mea.decompress_recording(rec12, downsample)
    assert res.sample_rate == pytest.approx(expected_sample_rate)

    # Test 2
    downsample = 100
    res = mea.decompress_recording(rec12, downsample)
    max_per_bucket = np.max(res.spikes)
    assert max_per_bucket == 3


def test_single_3brain_recording(
    stimulus_pattern, recorded_stimulus, response_data
):
    """
    Tests that:
        1. A recording can be loaded without errors (very basic checks).
        2. Filtering by cluster id works.
        3. Requesting non-existing clusters raises an error.
    """
    # Test 1
    # Setup
    expected_num_samples = 16071532
    # Test
    rec = mea._single_3brain_recording(
        "Chicken_17_08_21_Phase_00",
        stimulus_pattern,
        recorded_stimulus,
        response_data,
    )
    assert rec.num_sensor_samples == expected_num_samples
    assert rec.stimulus_pattern.shape[1] == mea.NUM_STIMULUS_LEDS
    # Note: the sampling frequency in the Pandas dataframe isn't as accurate
    # as the ELECTRODE_FREQ value.
    assert rec.sensor_sample_rate == pytest.approx(mea.ELECTRODE_FREQ)

    # Test 2
    # Setup
    cluster_ids = {13, 14, 202, 1485}
    # Test
    rec = mea._single_3brain_recording(
        "Chicken_17_08_21_Phase_00",
        stimulus_pattern,
        recorded_stimulus,
        response_data,
        cluster_ids,
    )
    assert len(rec.spike_events) == len(cluster_ids)
    assert len(rec.cluster_ids) == len(cluster_ids)
    assert set(rec.cluster_ids) == cluster_ids

    # Test 3
    # Setup
    partially_existing_clusters = {
        13,  # Exists
        14,  # Exists
        18,  # Does not exist
    }
    # Test
    with pytest.raises(ValueError):
        mea._single_3brain_recording(
            "Chicken_17_08_21_Phase_00",
            stimulus_pattern,
            recorded_stimulus,
            response_data,
            partially_existing_clusters,
        )


def test_decompress_stimulus():
    """
    Tests that:
        1. Decompression with no downsample works.
        2. Decompression with downsample works.
            - This test overlaps a bit with test_downsample_stimulus
        3. Invalid trigger start should cause an exception.
    """
    # Setup
    # fmt: off
    stimulus_pattern = np.array([
        [1, 1, 1, 1, 0, 0, 0, 0],
        [1, 1, 0, 0, 1, 1, 0, 0],
        [1, 0, 1, 0, 1, 0, 1, 0]]).T
    trigger_events = np.array([0, 4, 5, 7, 12])
    num_sensor_samples = 20
    expected_output = np.array([
    # idx: 0  1  2  3  4  5  6  7  8  9  10 11 12 13 14 15 16 17 18 19
    # val: 0  0  0  0  1  2  2  3  3  3  3  3  4  4  4  4  4  4  4  4
          [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
          [1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1],
          [1, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1]]).T
    expected_downsampled = np.array([
    # idx: 0     2     4     6     8     10    12    14    16    18
    # val: 0     0     1     2     3     3     4     4     4     4 
          [0.7, 1.1,   1,    1,    1,    1,  0.3,    0,    0,    0],
          [0.7, 1.1, 0.7,    0,    0,    0,  0.7,    1.1,  1,    1],
          [0.7,   1, 0.5,  0.7,    0,    0,  0.7,    1.1,  1,    1]]).T
    # fmt: on
    # Notice how the first downsampled value is pulled towards zero. This
    # is due to the "constant" option sent to scipy.resample_poly, which is
    # called by scipy.resample with "0" as the constant value. There are other
    # options, such as "mean" too. If we want a different constant value
    # or want "mean" padding, we should call scipy.resapmle_poly directly.

    # Test 1
    downsample = 1
    res = mea.decompress_stimulus(
        stimulus_pattern, trigger_events, num_sensor_samples, downsample
    )
    np.testing.assert_array_equal(res, expected_output)

    # Test 2
    downsample = 2
    res = mea.decompress_stimulus(
        stimulus_pattern, trigger_events, num_sensor_samples, downsample
    )
    # Test approx with numpy
    numpy.testing.assert_allclose(res, expected_downsampled, atol=0.1)

    # Test 3
    # Non-zero starting trigger is invalid.
    trigger_events_invalid = np.array([2, 4, 5, 7, 12])
    with pytest.raises(ValueError):
        mea.decompress_stimulus(
            stimulus_pattern,
            trigger_events_invalid,
            num_sensor_samples,
            downsample,
        )


def test_factors_sorted_by_count():
    """
    Tests that:
        1. A simple case:
            1. There are no duplicates.
            2. The factors are sorted by count.
        2. Setting a limit will cause the list to be truncated.
        3. If no factorizations meet the limit requirement, a fallback is
            returned.
        4. For a prime, the parameter is returned as is.
    """
    # Test 1
    num = 12
    expected = ((2, 2, 3), (2, 6), (3, 4), (12,))
    # Test 1
    res = mea.factors_sorted_by_count(num)
    assert set(res) == set(expected)

    # Test 2
    num = 12
    expected = ((2, 2, 3), (3, 4))
    res = mea.factors_sorted_by_count(num, limit=5)
    assert set(res) == set(expected)
    expected = ((2, 2, 3), (2, 6), (3, 4))
    res = mea.factors_sorted_by_count(num, limit=6)
    assert set(res) == set(expected), "The limit should be inclusive."

    # Test 3
    num = 2 * 2 * 13 * 13
    # Note: note how the result is not really ideal (13, 13, 4) would be better.
    expected = ((2, 2, 13, 13),)
    res = mea.factors_sorted_by_count(num, limit=5)
    assert set(res) == set(expected)

    # Test 4
    num = 89
    expected = ((89,),)
    res = mea.factors_sorted_by_count(num)
    assert res == expected


def test_downsample_stimulus():
    # Setup
    # fmt: off
    orig_signal = np.array([0, 0, 0, 0, 0, 0, 0, 0,
                            1, 1, 1, 1, 1, 1, 1, 1,
                            0, 0, 0, 0, 0, 0, 0, 0,
                            1, 1, 1, 1, 1, 1, 1, 1,
                            0, 0, 0, 0, 0, 0, 0, 0,
                            1, 1, 1, 1, 1, 1, 1, 1])
    # TODO: is this satisfactory?
    expected_decimate_by_2 = np.array(
            [0.012, -0.019,  0.029, -0.060,
             0.739,  1.085,  0.937,  1.081,
             0.249, -0.076,  0.058, -0.077,
             0.750,  1.077,  0.942,  1.077,
             0.250, -0.077,  0.058, -0.076,
             0.749,  1.081,  0.937,  1.085])
    expected_decimate_by_4 = np.array(
            [0.030, -0.065,  0.595,  1.148,
             0.364, -0.120,  0.620,  1.136,
             0.369, -0.120,  0.614,  1.148])
    # fmt: on

    # Test
    decimated_by_2 = mea.downsample_stimulus(orig_signal, 2)
    numpy.testing.assert_allclose(
        decimated_by_2, expected_decimate_by_2, atol=0.002
    )
    decimated_by_4 = mea.downsample_stimulus(orig_signal, 4)
    numpy.testing.assert_allclose(
        decimated_by_4, expected_decimate_by_4, atol=0.002
    )


def test_compress_spikes():
    """
    Tests that the basic spikes to spike-indexes function works.
    """
    # Setup
    spikes = np.array([0, 0, 1, 0, 0, 2, 3])
    expected_indexes = np.array([2, 5, 5, 6, 6, 6])

    # Test
    res = mea.compress_spikes(spikes)
    np.testing.assert_array_equal(res, expected_indexes)


def test_decompress_spikes1():
    # Setup
    downsample_by = 9
    num_sensor_samples = 123
    # fmt: off
    spike_times1 = np.array([8, 9, 30, 40, 50, 70, 80, 90, 100, 110])
    spike_times2 = np.array([0, 1, 8, 9, 10, 27, 30, 40, 50, 70, 80, 90, 100, 110])

    spike_counts1 = np.array([1, 1,  0,  1,  1,  1,  0,  1,  1,  0,  1,  1,  1,  0])
    spike_counts2 = np.array([3, 2,  0,  2,  1,  1,  0,  1,  1,  0,  1,  1,  1,  0])
    # fmt: on
    expected_output_len = math.ceil(num_sensor_samples / downsample_by)
    assert len(spike_counts1) == len(spike_counts2) == expected_output_len

    # Test 1
    spikes = mea.decompress_spikes(
        spike_times1, num_sensor_samples, downsample_by
    )
    assert np.array_equal(spikes, spike_counts1)

    # Test 2
    # Test the case where two spikes land in the same bucket.
    # There should *not* be an error thrown, even though two samples land in
    # the same bucket.
    spikes = mea.decompress_spikes(
        spike_times2, num_sensor_samples, downsample_by
    )
    assert np.array_equal(spikes, spike_counts2)


def test_decompress_spikes2(response_data):
    """
    Test decompress_spikes on actual recordings.

    Not much is actually checked though.
    """
    # Setup
    cluster_id = 36
    spikes_row = response_data.xs(
        (cluster_id, "Chicken_17_08_21_Phase_00"),
        level=("Cell index", "Recording"),
    ).iloc[0]
    spikes = spikes_row["Spikes"].compressed()

    assert spikes.shape == (2361,)
    num_sensor_samples = spikes[-1] + 1000

    # Test
    mea.decompress_spikes(spikes, num_sensor_samples)
    mea.decompress_spikes(spikes, num_sensor_samples, downsample_factor=18)


def test_spike_snippets():
    """
    Tests that the spike snippets are extracted correctly.

    Specifically, the following cases are considered:

        1. The spike needs no padding.
        2. The spike needs padding at the beginning.
        3. The spike happens in the first sample.
        4. The spike needs padding at the end.
        5. The spike happens in the last sample.
    """
    # Setup
    # -----
    # fmt: off
    # Test:   1, 2, 3, 4, 5
    spikes = [4, 1, 0, 6, 7]
    # The numbers in comments refer to the 5 tests below.
    stimulus = np.array(
        [
            [1, 1, 1, 1],  #     |  3
            [1, 1, 1, 1],  #     2  |
            [0, 1, 1, 1],  #  -  |  -
            [0, 0, 1, 1],  #  |  -
            [0, 0, 0, 1],  #  1       -
            [0, 0, 0, 0],  #  |       |  -
            [1, 0, 0, 0],  #  -       4  |
            [1, 1, 0, 0],  #          |  5
        ]              
    )
    # fmt: on
    total_len = 5
    pad = 2

    def _test(spike, expected_snippet):
        """Collects the repetitive call and test into a function."""
        snippets = mea.spike_snippets(
            stimulus,
            [spike],
            total_len,
            pad,
        )
        assert (
            len(snippets) == 1
        ), "One spike was given, expect only one snippet."
        snippet = snippets[0]
        numpy.testing.assert_array_equal(snippet, expected_snippet)

    # Test 1: case where no padding is needed.
    _test(
        spikes[0],
        expected_snippet=np.array(
            [
                [0, 1, 1, 1],
                [0, 0, 1, 1],
                [0, 0, 0, 1],
                [0, 0, 0, 0],
                [1, 0, 0, 0],
            ]
        ),
    )

    # Test 2: sample is near the beginning and needs padding.
    _test(
        spikes[1],
        expected_snippet=np.array(
            [
                [0, 0, 0, 0],
                [1, 1, 1, 1],
                [1, 1, 1, 1],
                [0, 1, 1, 1],
                [0, 0, 1, 1],
            ]
        ),
    )

    # Test 3: sample is _at_ the beginning and needs padding.
    _test(
        spikes[2],
        expected_snippet=np.array(
            [
                [0, 0, 0, 0],
                [0, 0, 0, 0],
                [1, 1, 1, 1],
                [1, 1, 1, 1],
                [0, 1, 1, 1],
            ]
        ),
    )

    # Test 4: sample is near the end and needs padding.
    _test(
        spikes[3],
        expected_snippet=np.array(
            [
                [0, 0, 0, 1],
                [0, 0, 0, 0],
                [1, 0, 0, 0],
                [1, 1, 0, 0],
                [0, 0, 0, 0],
            ]
        ),
    )

    # Test 5: sample is _at_ the end and needs padding.
    _test(
        spikes[4],
        expected_snippet=np.array(
            [
                [0, 0, 0, 0],
                [1, 0, 0, 0],
                [1, 1, 0, 0],
                [0, 0, 0, 0],
                [0, 0, 0, 0],
            ]
        ),
    )


def test_labeled_spike_snippets():
    """
    Create a fake response `DataFrame` and check that the spike snippets are
    calculated correctly.
    """
    # Setup
    snippet_len = 7
    snippet_pad = 2
    # Fake stimulus pattern.
    stimulus_pattern = np.array(
        [
            [0, 0, 0, 1],  # 0
            [0, 0, 1, 0],  # 1
            [0, 0, 1, 1],  # 2
            [0, 1, 0, 0],  # 3
            [0, 1, 0, 1],  # 4
            [0, 1, 1, 0],  # 5
            [0, 1, 1, 1],  # 6
            [1, 0, 0, 0],  # 7
            [1, 0, 0, 1],  # 8
            [1, 0, 1, 0],  # 9
            [1, 0, 1, 1],  # 10
            [1, 1, 0, 0],  # 11
        ]
    )

    # Fake response
    rec_name1 = "Chicken1"
    rec_name2 = "Chicken2"
    cluster_ids1 = [25, 40]
    cluster_ids2 = [17, 40]
    spike_events1 = np.array(
        [
            [1, 8],
            [6, 9],
        ],
        dtype=int,
    )
    spike_events2 = np.array(
        [
            [11],
            [9],
        ],
        dtype=int,
    )
    # Make SpikeRecordings.
    stimulus_events = np.arange(len(stimulus_pattern))
    recording1 = mea.CompressedSpikeRecording(
        rec_name1,
        stimulus_pattern,
        stimulus_events,
        spike_events1,
        cluster_ids1,
        sensor_sample_rate=1,
        num_sensor_samples=len(stimulus_pattern),
    )
    recording2 = mea.CompressedSpikeRecording(
        rec_name2,
        stimulus_pattern,
        stimulus_events,
        spike_events2,
        cluster_ids2,
        sensor_sample_rate=1,
        num_sensor_samples=len(stimulus_pattern),
    )

    # The expected snippets.
    expected_spike_snippets1 = np.array(
        [
            [
                [0, 0, 0, 0],  # pad
                [0, 0, 0, 0],  # pad
                [0, 0, 0, 0],  # pad
                [0, 0, 0, 1],  # 0
                [0, 0, 1, 0],  # 1 <-- spike
                [0, 0, 1, 1],  # 2
                [0, 1, 0, 0],  # 3
            ],
            [
                [0, 1, 0, 1],  # 4
                [0, 1, 1, 0],  # 5
                [0, 1, 1, 1],  # 6
                [1, 0, 0, 0],  # 7
                [1, 0, 0, 1],  # 8 <- spike
                [1, 0, 1, 0],  # 9
                [1, 0, 1, 1],  # 10
            ],
            [
                [0, 0, 1, 1],  # 2
                [0, 1, 0, 0],  # 3
                [0, 1, 0, 1],  # 4
                [0, 1, 1, 0],  # 5
                [0, 1, 1, 1],  # 6 <- spike
                [1, 0, 0, 0],  # 7
                [1, 0, 0, 1],  # 8
            ],
            [
                [0, 1, 1, 0],  # 5
                [0, 1, 1, 1],  # 6
                [1, 0, 0, 0],  # 7
                [1, 0, 0, 1],  # 8
                [1, 0, 1, 0],  # 9 <- spike
                [1, 0, 1, 1],  # 10
                [1, 1, 0, 0],  # 11
            ],
        ]
    )
    expected_spike_snippets2 = np.array(
        [
            [
                [1, 0, 0, 0],  # 7
                [1, 0, 0, 1],  # 8
                [1, 0, 1, 0],  # 9
                [1, 0, 1, 1],  # 10
                [1, 1, 0, 0],  # 11 <- spike
                [0, 0, 0, 0],  # pad
                [0, 0, 0, 0],  # pad
            ],
            [
                [0, 1, 1, 0],  # 5
                [0, 1, 1, 1],  # 6
                [1, 0, 0, 0],  # 7
                [1, 0, 0, 1],  # 8
                [1, 0, 1, 0],  # 9 <- spike
                [1, 0, 1, 1],  # 10
                [1, 1, 0, 0],  # 11
            ],
        ]
    )
    expected_cluster_ids1 = np.array([25, 25, 40, 40])
    expected_cluster_ids2 = np.array([17, 40])

    # Test 1 (rec_name1)
    spike_snippets, cluster_ids = mea.labeled_spike_snippets(
        recording1,
        snippet_len,
        snippet_pad,
    )
    for idx, (spwin, cluster_ids) in enumerate(
        zip(spike_snippets, cluster_ids)
    ):
        np.testing.assert_equal(spwin, expected_spike_snippets1[idx])
        np.testing.assert_equal(cluster_ids, expected_cluster_ids1[idx])

    # Test 2 (rec_name2)
    spike_snippets, cluster_ids = mea.labeled_spike_snippets(
        recording2, snippet_len, snippet_pad
    )
    for idx, (spwin, cluster_ids) in enumerate(
        zip(spike_snippets, cluster_ids)
    ):
        np.testing.assert_equal(spwin, expected_spike_snippets2[idx])
        np.testing.assert_equal(cluster_ids, expected_cluster_ids2[idx])



from math import dist
import pytest
import retinapy.spikedistancefield as sdf
import numpy as np
import torch


@pytest.fixture
def spike_data1():
    spike_batch = np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, 1, 0, 0, 0, 1, 0, 0],
            [0, 1, 1, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ]
    )

    MAX_DIST = 6
    M = MAX_DIST  # Used to make the below array literal tidier.

    after_field = np.array(
        [
            [0, 1, 2, 3, 4, 5, M, M, M, 0],
            [M, M, 0, 1, 2, 3, 4, 5, M, M],
            [M, M, M, M, M, M, M, 0, 1, 2],
            [M, 0, 1, 0, 1, 2, 3, 0, 1, 2],
            [M, 0, 0, 0, 1, 2, 3, 4, 5, M],
            [M, M, M, M, M, M, M, M, M, M],
        ]
    )

    before_field = np.array(
        [
            [0, M, M, M, 5, 4, 3, 2, 1, 0],
            [2, 1, 0, M, M, M, M, M, M, M],
            [M, M, 5, 4, 3, 2, 1, 0, M, M],
            [1, 0, 1, 0, 3, 2, 1, 0, M, M],
            [1, 0, 0, 0, M, M, M, M, M, M],
            [M, M, M, M, M, M, M, M, M, M],
        ]
    )

    return (spike_batch, MAX_DIST, before_field, after_field)


@pytest.mark.skip(reason="Broken. See: GitHub issue #2.")
def test_bi_distance_field(spike_data1):
    spike_batch, max_dist, before_field, after_field = spike_data1
    for i in range(spike_batch.shape[0]):
        dist_before, dist_after = sdf.bi_distance_field(
            spike_batch[i], max_dist, max_dist
        )
        assert np.array_equal(dist_before, before_field[i])
        assert np.array_equal(dist_after, after_field[i])


@pytest.fixture
def spike_data2():
    spike_batch = np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, 1, 0, 0, 0, 1, 0, 0],
            [0, 1, 1, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ]
    )  # just for reference

    MAX_DIST = 100
    M = MAX_DIST  # Used to make the below array literal tidier.

    after_field = np.array(
        [
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 0],
            [M, M, 0, 1, 2, 3, 4, 5, 6, 7],
            [M, M, M, M, M, M, M, 0, 1, 2],
            [M, 0, 1, 0, 1, 2, 3, 0, 1, 2],
            [M, 0, 0, 0, 1, 2, 3, 4, 5, 6],
            [M, M, M, M, M, M, M, M, M, M],
        ]
    )

    before_field = np.array(
        [
            [0, 8, 7, 6, 5, 4, 3, 2, 1, 0],
            [2, 1, 0, M, M, M, M, M, M, M],
            [7, 6, 5, 4, 3, 2, 1, 0, M, M],
            [1, 0, 1, 0, 3, 2, 1, 0, M, M],
            [1, 0, 0, 0, M, M, M, M, M, M],
            [M, M, M, M, M, M, M, M, M, M],
        ]
    )

    spike_counts = np.array([2, 1, 1, 3, 3, 0])

    return (spike_counts, MAX_DIST, before_field, after_field)


@pytest.mark.skip(reason="Broken. See: GitHub issue #1.")
def test_count_inference_from_bi_df(spike_data2):
    spike_counts, max_dist, before_field, after_field = spike_data2
    before_field = torch.from_numpy(before_field).to(dtype=torch.float32)
    after_field = torch.from_numpy(after_field).to(dtype=torch.float32)
    for i in range(len(spike_counts)):
        num_spikes = sdf.count_inference_from_bi_df(
            before_field[i],
            after_field[i],
            lhs_spike=-max_dist,
            rhs_spike=len(before_field[i]) + max_dist - 1,
            spike_pad=1,
            target_interval=(0, len(before_field[i])),
            max_num_spikes=max_dist,
        )
        assert num_spikes == spike_counts[i]

@pytest.mark.skip(reason="Broken. See: GitHub issue #1.")
def test_count_inference_from_bi_df2(spike_data2):
    spike_counts, max_dist, before_field, after_field = spike_data2
    before_field = torch.from_numpy(before_field).to(dtype=torch.float32)
    after_field = torch.from_numpy(after_field).to(dtype=torch.float32)
    num_spikes = sdf.count_inference_from_bi_df2(
        before_field,
        after_field,
        torch.full((6,1), -max_dist),
        torch.full((6,1), before_field.shape[1] + max_dist - 1),
        spike_pad=1,
        target_interval=(0, before_field.shape[1]),
    )
    assert np.array_equal(num_spikes.numpy(), spike_counts)


@pytest.fixture
def distance_field_data():
    MAX_DIST = 100
    M = MAX_DIST  # Used to make the below array literal tidier.
    spike_batch = np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, 1, 0, 0, 0, 1, 0, 0],
            [0, 1, 1, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ]
    )
    distance_field = np.array(
        [
            [0, 1, 2, 3, 4, 4, 3, 2, 1, 0],
            [2, 1, 0, 1, 2, 3, 4, 5, 6, 7],
            [7, 6, 5, 4, 3, 2, 1, 0, 1, 2],
            [1, 0, 1, 0, 1, 2, 1, 0, 1, 2],
            [1, 0, 0, 0, 1, 2, 3, 4, 5, 6],
            [M, M, M, M, M, M, M, M, M, M],
        ]
    )
    return MAX_DIST, spike_batch, distance_field


def test_distance_field(distance_field_data):
    M, spikes, dist_fields = distance_field_data
    for spike, known_df in zip(spikes, dist_fields):
        dist_field = sdf.distance_field(spike, M)
        dist_field_cpu = sdf.distance_field2(spike, M)
        assert np.array_equal(known_df, dist_field)
        assert np.array_equal(dist_field_cpu, dist_field)


@pytest.fixture
def spike_interval_data():
    MAX_COUNT = 100
    M = MAX_COUNT  # Used to make the below array literal tidier.
    spike_batch = np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, 1, 0, 0, 0, 1, 0, 0],
            [0, 1, 1, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ]
    )
    spike_intervals = np.array(
        [
            [0, 8, 8, 8, 8, 8, 8, 8, 8, 0],
            [M, M, 0, M, M, M, M, M, M, M],
            [M, M, M, M, M, M, M, 0, M, M],
            [M, 0, 1, 0, 3, 3, 3, 0, M, M],
            [M, 0, 0, 0, M, M, M, M, M, M],
            [M, M, M, M, M, M, M, M, M, M],
        ]
    )
    return MAX_COUNT, spike_batch, spike_intervals

def test_spike_interval(spike_interval_data):
    M, spikes, spike_intervals = spike_interval_data
    for spike, known_si in zip(spikes, spike_intervals):
        si = sdf.spike_interval(spike, M)
        assert np.array_equal(known_si, si)


def test_mle_inference_from_df(distance_field_data):
    M, spikes, dist_fields = distance_field_data
    for spike, dist_field in zip(spikes, dist_fields):
        num_spikes = sdf.mle_inference_from_df(torch.Tensor(dist_field), 
                                               lhs_spike=-M,
                                               rhs_spike=len(spike) + M - 1,
                                               spike_pad=1,
                                               max_clamp=M*2,
                                               max_num_spikes=5,
                                               resolution=1)





from typing import Optional

import numpy as np

import plotly
import plotly.graph_objects as go
import plotly.subplots
import retinapy
import retinapy.dataset
import retinapy.mea as mea
import retinapy.spikeprediction as sp
import torch
import sklearn.manifold
from typing import Union, Iterable
import pathlib
import PIL.Image
import logging
import functools
import concurrent.futures
import scinot

"""
This module is starting out as somewhere to create the most commonly figures 
needed in notebooks. Going forward, it's hoped that some of the plots or 
images created here can be recorded automatically when training a model.
"""

_logger = logging.getLogger(__name__)


def create_title(title: str, subtitle: Optional[str]) -> str:
    """
    Returns a styled title string for use in plotly.

    Subtitle is optional.
    """
    if subtitle is None:
        res = f"{title}"
    else:
        res = f'{title}<br><span style="font-size:90%; color:grey">{subtitle}</span>'
    return res


def create_axis_title(title: str, units: str) -> str:
    """Returns a styled title string for use in plotly."""
    return f"{title}<span style='font-size:90%;white-space:pre;color:grey'>  ({units})</span>"


def default_fig_layout():
    """Some default fig options.

    For quite a few of these, like margin, I don't quite understand their
    full implications. This is part of the reason it's useful to move them
    here, as they will likely change over time.
    """
    res = {
        "title_x": 0.5,
        "title_pad": dict(l=0, r=0, b=0, t=0),
        # The title and xaxis label fit within the margin, so it needs
        # to be big enough to fit them. Top margin of 50 or so works for a
        # title with no subtitle. 70 works for a title with a subtitle.
        # 50 is okay for bottom, but it's a little too tight in Jupyter Lab,
        # so making it a tad bigger (60).
        "margin": {"l": 0, "r": 0, "t": 50, "b": 60, "pad": 0},
    }
    return res


def kernel(
    kernel: np.ndarray,
    t_0: int,
    bin_duration_ms,
    vline: Optional[int] = None,
    title=None,
    subtitle=None,
):
    """
    Args:
        kernel: The kernel to plot. This is a 2D array with shape
            (n_channels, n_timesteps). There can be 1-4 channels.
        t_0: the index of the bin that corresponds to t=0.
    """
    fig = go.Figure()
    xs = (np.arange(kernel.shape[0]) - t_0) * bin_duration_ms
    if vline is not None:
        fig.add_vline(
            x=-100,
            line_width=2,
            line_dash="dot",
            line_color="grey",
            annotation_text="-100ms",
            annotation_position="bottom right",
        )
    for idx, stim in enumerate(mea.stimuli):
        fig.add_trace(
            go.Scatter(
                x=xs,
                y=kernel[idx],
                name=stim.name,
                mode="lines",
                line=dict(width=2),
            )
        )
        fig.add_trace(
            go.Scatter(
                x=xs,
                y=kernel[:, idx],
                line_color=stim.display_hex,
                mode="lines",
            )
        )
    fig.update_layout(default_fig_layout())
    fig.update_layout(
        margin=dict(l=1, r=1, b=1, t=25, pad=10),
        yaxis_fixedrange=True,
        showlegend=False,
        title="Kernel",
        xaxis={"title": "time (ms), with spike at 0"},
        yaxis={"title": "Stimulus", "range": [0, 1]},
    )
    return fig


class KernelPlots:
    """Creates and saves kernel plots as images to disk.

    Motivation:
    This class was created to speed up the process of creating kernel plots when
    they were needed in bulk for hover-over tooltips.
    """

    FILE_PATH_PATTERN = "{rec_name}_c{cluster_id}.png"
    OUT_DIR = "kernel_plots"

    def __init__(self, img_dir: Union[str, pathlib.Path]):
        self.img_dir = pathlib.Path(img_dir)
        if not self.img_dir.is_dir():
            raise ValueError(
                "img_dir must be a directory (with the saved kernel plots)."
            )
        self.cache = {}

    def get(self, rec_name, cluster_id):
        if (rec_name, cluster_id) in self.cache:
            return self.cache[(rec_name, cluster_id)]
        img = PIL.Image.open(
            self.img_dir
            / self.FILE_PATH_PATTERN.format(
                rec_name=rec_name, cluster_id=cluster_id
            )
        )
        self.cache[(rec_name, cluster_id)] = img
        return img

    @staticmethod
    def _generate_single(
        ds_rec, c_idx, snippet_len, snippet_pad, mini, out_dir
    ):
        snippets = mea.spike_snippets(
            ds_rec.stimulus,
            mea.compress_spikes(ds_rec.spikes[:, c_idx]),
            snippet_len,
            snippet_pad,
        )
        # Kernel and spike counts.
        c_id = ds_rec.cluster_ids[c_idx]
        ker = snippets.mean(axis=0)
        num_spikes = len(snippets)
        # Make figure.
        bin_duration_ms = 1000 / ds_rec.sample_rate
        t_0 = snippet_len - snippet_pad
        fig = kernel(ker, t_0=t_0, bin_duration_ms=bin_duration_ms)
        if mini:
            fig.update_layout(
                {
                    "yaxis": {"visible": False},
                    "xaxis": {"title": None},
                    "title": {
                        "text": (
                            f'<span style="font-size:75%">{ds_rec.name}</span><br>'
                            f'<span style="font-size:95%">c{c_id}</span> '
                            '<span style="font-size:85%">'
                            f'({scinot.format(num_spikes, 2)} spikes</span>)' 
                        ),
                    },
                    "width": 200,
                    "height": 200,
                }
            )
        else:
            fig.update_layout(
                {
                    "title": {
                        "text": create_title(
                            "STA kernel",
                            f"rec: {ds_rec.name}, cluster: {c_id}",
                        )
                    }
                }
            )
        # Save figure.
        fig.write_image(
            out_dir
            / KernelPlots.FILE_PATH_PATTERN.format(
                rec_name=ds_rec.name, cluster_id=c_id
            )
        )

    @classmethod
    def generate(
        cls,
        recs: Iterable[mea.SpikeRecording],
        snippet_len: int,
        snippet_pad: int,
        out_dir: Union[str, pathlib.Path],
        mini=False,
        num_workers=4,
    ) -> "KernelPlots":
        """
        Args:
            recs: The recordings to generate the kernel plots for.
            snippet_len: The length of the snippet to use for the kernel.
            snippet_pad: The number of padding bins to use after the spike.
            out_dir: The directory to save the plots to.
            mini: If True, make the plots smaller by removing axes and making
                the images smaller.
        """
        # Create outer directory.
        out_dir = pathlib.Path(out_dir) / cls.OUT_DIR
        exists_and_not_empty = out_dir.is_dir() and len(list(out_dir.iterdir()))
        if exists_and_not_empty:
            raise ValueError(
                f"out_dir already exists and has contents ({out_dir})"
            )
        out_dir.mkdir(parents=False, exist_ok=True)

        _logger.info(f"Creating kernel plots in {out_dir}")
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=num_workers
        ) as executor:
            _func = functools.partial(
                cls._generate_single,
                snippet_len=snippet_len,
                snippet_pad=snippet_pad,
                mini=mini,
                out_dir=out_dir,
            )
            futures = {}
            for ds_rec in recs:
                for c_idx in range(len(ds_rec.cluster_ids)):
                    f = executor.submit(_func, ds_rec=ds_rec, c_idx=c_idx)
                    futures[f] = (ds_rec, c_idx)
            # Wait for all futures to complete.
            for future in concurrent.futures.as_completed(futures):
                ds_rec, c_idx = futures[future]
                try:
                    future.result()
                except Exception as err:
                    _logger.error(
                        f"Error creating kernel plot for cluster {c_idx} of "
                        f"recording {ds_rec.name}: {err}"
                    )
        _logger.info(f"Finished creating kernel plots")
        return KernelPlots(out_dir)


def stimulus_fig(
    stimulus: np.ndarray, start_ms: float = 0, bin_duration_ms: float = 1.0
):
    """
    A figure to display a stimulus.

    Args:
        stimulus: the 4 color stimulus, with shape (n_channels=4, n_timesteps)
    """
    fig = plotly.subplots.make_subplots(
        rows=2,
        cols=1,
        shared_xaxes=True,
        vertical_spacing=0.05,
        row_width=[0.2, 0.8],
        x_title=create_axis_title("time", "ms"),
    )
    xs = np.arange(stimulus.shape[1]) * bin_duration_ms + start_ms
    # RGBU
    for idx, stim in enumerate(mea.stimuli):
        fig.append_trace(
            go.Scatter(
                x=xs,
                y=stimulus[idx, :],
                line_color=stim.display_hex,
                name=f"{stim.wavelength} nm",
                mode="lines",
            ),
            row=1,
            col=1,
        )
    # sum
    fig.append_trace(
        go.Scatter(
            x=xs,
            y=stimulus.sum(axis=0),
            line_color="black",
            name="sum",
            mode="lines",
        ),
        row=2,
        col=1,
    )
    fig.update_yaxes(
        {"tickmode": "array", "tickvals": [-5, 0, 5]}, row=2, col=1
    )
    # Set a default layout.
    fig.update_layout(default_fig_layout())
    fig.update_layout(
        {
            "title": {"text": "Stimulus"},
            # The title and xaxis label fit within the margin, so it needs
            # to be big enough to fit them.
            "yaxis": {
                "title": {"text": create_axis_title("stimulus", "no units")},
                "fixedrange": True,
                # The range is dependent on the agumentation.
                "range": [-2, 2],
                "tickmode": "array",
                "tickvals": [-1, 0, 1],
            },
        }
    )
    return fig


def distfield_fig(
    actual: np.ndarray,
    pred: np.ndarray,
    start_ms: float = 0,
    bin_duration_ms: float = 1.0,
    stride_bins: int = 0,
    log_space: bool = False,
):
    """
    A figure containing 1 or multiple line charts comparing distance fields.

    Args:
        stride: the jump in the x-axis origin label between each subplot. This
            is used to plot multiple times where each subplot starts say
            50 ms after the previous one.
    """
    if actual.shape != pred.shape:
        raise ValueError(
            "Both distfield inputs must have the same shape. Got "
            f"{actual.shape} and {pred.shape}."
        )
    # Expand dims if only one distfield is given (single or multiple distfields
    # are accepted).
    if len(actual.shape) == 1:
        actual = actual.reshape(1, -1)
        pred = pred.reshape(1, -1)
    num_rows = len(actual)
    # Initialize the figure, in its subplot glory.
    fig = plotly.subplots.make_subplots(
        rows=num_rows,
        cols=1,
        shared_xaxes=False,
        # Default vertical spacing is 0.3/num_rows
        vertical_spacing=0.5 / num_rows,
        # Make the plots have the same y-range.
        shared_yaxes="columns",
        x_title=create_axis_title("time", "ms"),
        y_title=(
            create_axis_title("log time to nearest spike", "log(ms)")
            if log_space
            else create_axis_title("time to nearest spike", "ms")
        ),
    )
    # Create the scatters.
    stride_ms = stride_bins * bin_duration_ms
    num_bins = actual.shape[1]
    for r in range(num_rows):
        start = start_ms + r * stride_ms
        xs = np.arange(num_bins) * bin_duration_ms + start
        showlegend = r == 0
        scatter_actual = go.Scatter(
            x=xs,
            y=actual[r, :],
            name="actual",
            mode="lines",
            line_color="tomato",
            showlegend=showlegend,
            legendgroup="actual",
        )
        scatter_pred = go.Scatter(
            x=xs,
            y=pred[r, :],
            name="pred",
            mode="lines",
            line_color="gray",
            showlegend=showlegend,
            legendgroup="pred",
        )
        fig.add_trace(scatter_actual, row=r + 1, col=1)
        fig.add_trace(scatter_pred, row=r + 1, col=1)
    # Determine the y-axis scale. These are defaults, free to be changed.
    if log_space:
        yaxis_range = [-4, 4]
    else:
        yaxis_range = [-1, 600]
    # Create a title.
    total_len = actual.shape[1] * bin_duration_ms
    title_start = "Log distance fields" if log_space else "Distance fields "
    stride_str = (
        f"strided by {(stride_bins)} bins ({stride_ms:.1f} ms)"
        if stride_bins > 0
        else "no stride specified"
    )

    title_str = create_title(
        f"{title_start}, actual and predicted",
        f"{num_rows} snippets, "
        f"{num_bins} bins x {bin_duration_ms:.1f} ms/bin "
        f"= {total_len:.1f} ms each, {stride_str} ",
    )

    height = 140 * num_rows
    fig.update_layout(default_fig_layout())
    fig.update_layout(
        {
            "height": height,
            "yaxis": {
                "range": yaxis_range,
                "fixedrange": True,
            },
            "title": {"text": title_str},
            # Need left margin to fit the shared y-axis title, and some extra
            # on top for the subtitle.
            "margin": {"l": 70, "t": 70},
        }
    )
    return fig


def distfield_model_in_out(
    stimulus: np.ndarray,
    in_spikes: np.ndarray,
    target_dist: np.ndarray,
    model_out: np.ndarray,
    pos_enc: Optional[np.ndarray] = None,
    out_spikes: Optional[np.ndarray] = None,
    start_ms=0,
    bin_duration_ms=1.0,
    cluster_label=None,
):
    """A figure to view the inputs and outputs of a distfield model.

    Create a figure that has the stimulus on top, and the input spikes and
    the pair of distfields on the bottom (actual, pred). Roughly formatted
    like so:

    +--------------------+
    |      stimulus      | (includes the position encoding, if any)
    +--------------------+
    | spikes      | dist |
    +--------------------+

    """
    # The figure will be built in 4 steps:
    #   1. The 4-channel stimuli that spans the whole x-axis.
    #       1.1 Draw a 5th line for the position encoding, if any.
    #   2. The sum of the 4-channel stimuli (in black). This line is plotted
    #      below the 4-channel stimuli and also spans the whole x-axis.
    #   3. The model output distance field and the true (target) distance field.
    #      These lines occupy a small time-slice on the right, typically around
    #      50-200 ms depending on the model.
    #   4. The input spike data, shown as vertical lines on the same sub-plot
    #      as the distance fields.

    # It would be nice to reuse the functionality of the above two functions;
    # however, there are just enough differences for it to be not so easy.

    fig = plotly.subplots.make_subplots(
        rows=6,
        cols=1,
        shared_xaxes=True,
        vertical_spacing=0.02,
        #row_width=[0.15, 0.15, 0.15, 0.15, 0.15, 0.45],
        x_title=create_axis_title("time", "ms"),
    )

    # 1. The 4-channel stimulus data.
    stim_len = stimulus.shape[1]
    xs = np.arange(stim_len) * bin_duration_ms + start_ms
    # RGBU
    for idx, stim in enumerate(mea.stimuli):
        fig.append_trace(
            go.Scatter(
                x=xs,
                y=stimulus[idx, :],
                line_color=stim.display_hex,
                name=f"{stim.wavelength} nm",
                mode="lines",
            ),
            row=idx+1,
            col=1,
        )
        fig.update_yaxes(
            {"tickmode": "array", "tickvals": [-1, 0, 1], "fixedrange": True,
            "title_text": ['R', 'G', 'B', 'UV'][idx], "title_standoff": 0 },
            row=idx+1,
            col=1,
        )
    if pos_enc is not None:
        fig.append_trace(
            go.Scatter(
                x=xs,
                y=pos_enc,
                line_color="black",
                name="pos enc",
                mode="lines",
            ),
            row=5,
            col=1,
        )

    # 2. The summed stimulus data.
    fig.append_trace(
        go.Scatter(
            x=xs,
            y=stimulus.sum(axis=0),
            line_color="black",
            name="sum",
            mode="lines",
        ),
        row=5,
        col=1,
    )
    fig.update_yaxes(
        {"tickmode": "array", "tickvals": [-5, 0, 5], "fixedrange": True,
        "range": [-6, 6]},
        row=5,
        col=1,
    )

    # 3. The model output distance field and the target (actual).
    # Quite a dependency (on DistFieldDataset) creeping in here! But we need
    # to know what mask value the dataset is using. So it is necessary. I just
    # find it interesting to note the dependency. Can there be a project wide
    # mask value? Probably not a useful idea, is my guess.
    mask_start_idx = retinapy.dataset.DistFieldDataset.mask_start(in_spikes)
    xs_dist = xs[mask_start_idx:]
    if len(xs_dist) != len(model_out):
        raise ValueError(
            "The spike mask isn't the same lengths as the model output. "
            f"Got ({len(xs_dist)}) and ({len(model_out)})"
        )
    if mask_start_idx + len(model_out) != stim_len:
        raise ValueError(
            "The stimulus length doesn't allign with the model " "output."
        )
    fig.append_trace(
        go.Scatter(
            x=xs_dist,
            y=model_out,
            name="pred",
            mode="lines",
            line_color="gray",
        ),
        col=1,
        row=6,
    )
    fig.append_trace(
        go.Scatter(
            x=xs_dist,
            y=target_dist,
            name="actual",
            mode="lines",
            line_color="tomato",
        ),
        col=1,
        row=6,
    )
    fig.update_yaxes({"range": [-4, 4]}, row=6, col=1)
    # Cover the whole output region with a vrect to separate the regions.
    fig.add_vrect(
        x0=xs[mask_start_idx],
        x1=xs[-1],
        fillcolor="aqua",
        opacity=0.25,
        col=1,
        row=6,
        line_width=0,
        layer="below",
    )

    # 4.1. The vertical lines marking the input spikes.
    index_of_spikes = np.flatnonzero(in_spikes > 0)
    for idx in index_of_spikes:
        spike_loc = start_ms + idx * bin_duration_ms
        fig.add_vline(
            x=spike_loc, line_color="tomato", line_dash="dot", row=6, col=1
        )

    # 4.2. The vertical lines marking the output spikes.
    if out_spikes is not None:
        index_of_spikes = np.flatnonzero(out_spikes > 0) + mask_start_idx
        for idx in index_of_spikes:
            spike_loc = start_ms + idx * bin_duration_ms
            fig.add_vline(
                x=spike_loc, line_color="gray", line_dash="dot", row=6, col=1
            )

    # Set a default layout.
    fig.update_layout(default_fig_layout())
    # Create title
    main_title_str = "Spike distance model input-output"
    if cluster_label is not None:
        main_title_str += f" ({cluster_label})"
    title = create_title(
        main_title_str,
        "Output region is highlight blue. Spikes are red dashed lines.",
    )
    fig.update_layout(
        {
            "title": {"text": title},
            "height": 600,
            "margin": {"t": 70, "l": 100},
        }
    )
    fig.add_annotation(text=create_axis_title("RGB-UV stimulus (normalized)", "no units"),
                  xref="paper", yref="paper", #font_color="grey",
                  x=-0.1, y=0.8, textangle=-90, showarrow=False)
    fig.add_annotation(text="<span style=''>spike distance<br>(log, normalized)<br>"
                       "<span style='font-size:90%;whitespace:pre;color:grey'>log(ms)</span>",
                  xref="paper", yref="paper", #font_color="grey",
                  x=1.15, y=0.0, showarrow=False)
    return fig


def latent_tsne_fig(
    rec_ids: np.ndarray,
    cluster_ids: np.ndarray,
    zs: np.ndarray,
    use_label=False,
    perplexity=25,
):
    """Arbitrary dimensional latent points plotted in 2D using t-SNE.

    Args:
        perplexity: "perplexity" parameter for t-SNE. This option makes a big
            difference to the output. The default has been set based on some
            quick trial and error and guesswork, while playing with the latent
            spike of a 40D transformer VAE model.
    """

    # 1. Run t-SNE.
    z_2d = sklearn.manifold.TSNE(
        n_components=2,
        early_exaggeration=12,  # default = 12
        learning_rate="auto",
        n_iter=5000,  # default = 1000
        min_grad_norm=1e-7,  # default = 1e-7
        init="random",  # default = 'random', alternative 'pca'
        angle=0.6,  # default = 0.5. Higher is faster.
        n_jobs=20,  # default = 1
        perplexity=25,
    ).fit_transform(zs)

    # 1. Gather the data to plot. We will actually do that here, so this is
    # quite a proactive plotting function.
    fig = go.Figure()
    labels = [
        f"({r_idx}, {c_id})" for (r_idx, c_id) in zip(rec_ids, cluster_ids)
    ]
    mode = "markers+text" if use_label else "markers"
    scatter = go.Scatter(
        x=z_2d[:, 0],
        y=z_2d[:, 1],
        text=labels,
        textposition="bottom center",
        mode=mode,
    )
    fig.add_trace(scatter)
    fig.update_layout(
        {
            "title": {
                "text": f"Latent space, (t-SNE reduction from {zs.shape[1]} dimensions)"
            },
        }
    )
    return fig


def latent2d_fig(
    rec_ids: np.ndarray,
    cluster_ids: np.ndarray,
    z_xs: np.ndarray,
    z_ys: np.ndarray,
    use_label=False,
):
    """Basic 2D plot of 2D latent variables."""
    # 1. Gather the data to plot. We will actually do that here, so this is
    # quite a proactive plotting function.
    fig = go.Figure()
    labels = [
        f"({r_idx}, {c_id})" for (r_idx, c_id) in zip(rec_ids, cluster_ids)
    ]
    mode = "markers+text" if use_label else "markers"
    scatter = go.Scatter(
        x=z_xs,
        y=z_ys,
        text=labels,
        textposition="bottom center",
        mode=mode,
    )
    fig.add_trace(scatter)
    fig.update_layout(
        {
            "title": {"text": "Latent space, z"},
            "xaxis": {"range": [-3, 3]},
            "yaxis": {"range": [-3, 3]},
        }
    )
    return fig



import pytest
import retinapy.mea as mea
import retinapy.vis as vis


def test_KernelPlots(five_dc_recs, tmp_path):
    """Tests KernelPlots class.

    Tests that:
        1. Kernel plots generate() function can be called without error.
        2. Loading a kernel plot img after calling generate() works.
    """
    # Setup
    out_dir = tmp_path / "test_KernelPlots"
    out_dir.mkdir()
    # Test
    # 1. Generate plots.
    kernel_plots = vis.KernelPlots.generate(
            five_dc_recs,
            snippet_len=500,
            snippet_pad=50,
            out_dir=out_dir,
            mini=True,
            num_workers=100)
    # 2. Load all expected plots.
    for rec in five_dc_recs:
        for cluster_id in rec.cluster_ids:
            plot_img = kernel_plots.get(rec.name, cluster_id)
            assert plot_img is not None


    






import retinapy.mea as mea
import retinapy.spikedistancefield as sdf
import numpy as np
import torch
import bisect
import math
from typing import Tuple, Optional, List, Iterable


def _num_snippets(num_timesteps: int, snippet_len: int, stride: int) -> int:
    """
    Returns the number of snippets that can be extracted from a recording with
    the given number of timesteps, snippet length, and stride.
    """
    # The most likely place for insidious bugs to hide is here. So try two
    # ways.
    # 1. Find the last strided timestep, then divide by stride.
    last_nonstrided = num_timesteps - snippet_len
    last_strided_timestep = last_nonstrided - (last_nonstrided % stride)
    num_snippets1 = last_strided_timestep // stride + 1
    # 2. Same deal, but mindlessly copying the formula from Pytorch docs.
    # https://pytorch.org/docs/stable/generated/torch.nn.Conv1d.html
    # Padding is 0, and dilation is 1.
    num_snippets2 = math.floor(
        (num_timesteps - (snippet_len - 1) - 1) / stride + 1
    )
    assert num_snippets1 == num_snippets2, (
        f"Strided snippet calculation is wrong. {num_snippets1} != "
        f"{num_snippets2}."
    )
    return num_snippets1


class SnippetDataset(torch.utils.data.Dataset):
    def __init__(
        self, recording: mea.SpikeRecording, snippet_len: int, stride: int = 1
    ):
        if snippet_len > len(recording):
            raise ValueError(
                f"Snippet length ({snippet_len}) is larger than "
                f"the recording length ({len(recording)})."
            )
        self.recording = recording
        self.snippet_len = snippet_len
        self.stride = stride
        self.num_clusters = len(recording.cluster_ids)
        self.num_timesteps = len(recording) - snippet_len + 1
        assert (
            self.num_timesteps > 0
        ), "Snippet length is longer than the recording."
        self.num_strided_timesteps = _num_snippets(
            len(recording), snippet_len, stride
        )

    def __len__(self):
        """
        Calculates the number of samples in the dataset.
        """
        return self.num_strided_timesteps * self.num_clusters

    def _decode_index(self, index: int) -> Tuple[int, int]:
        """
        Decodes the index into the timestep and cluster id.

        The data is effectively a 2D array with dimensions (time, cluster).
        The index is the flattened index of this array, and so the timestep
        increases as the index increases and wraps to the next cluster id when
        it reaches the end of the recording.
        """
        timestep_idx = self.stride * (index % self.num_strided_timesteps)
        cluster_idx = index // self.num_strided_timesteps
        return timestep_idx, cluster_idx

    def __getitem__(self, idx):
        """
        Returns the snippet at the given index.
        """
        start_time_idx, cluster_idx = self._decode_index(idx)
        end_time_idx = start_time_idx + self.snippet_len
        assert end_time_idx <= len(self.recording)
        rec = self.recording.stimulus[start_time_idx:end_time_idx].T
        spikes = self.recording.spikes[start_time_idx:end_time_idx, cluster_idx]
        res = {"stimulus": rec, "spikes": spikes, "cluster_id": cluster_idx}
        return res

    @property
    def sample_rate(self):
        return self.recording.sample_rate


class SpikeCountDataset(torch.utils.data.Dataset):
    """
    Dataset that pairs a stimulus+spike history with a future spike count.

    The spike count is for a configurable duration after the end of the
    history snippet.

    This (X,y) style dataset is intended to for basic comparison between
    different spike count models.
    """

    def __init__(
        self, recording: mea.SpikeRecording, input_len: int, output_len: int
    ):
        self.output_len = output_len
        self.input_len = input_len
        self.total_len = input_len + output_len
        self.recording = recording
        self.ds = SnippetDataset(recording, self.total_len)

    def __len__(self):
        """
        Calculates the number of samples in the dataset.

        There will be one sample for every timestep in the recording.
        """
        res = len(self.recording) - self.total_len + 1
        assert res > 0, "Snippet length is longer than the recording."
        return res

    def __getitem__(self, idx):
        """
        Returns the (X,y) sample at the given index.

        Index is one-to-one with the timesteps in the recording.
        """
        sample = self.ds[idx]
        rec = sample["stimulus"]
        spikes = sample["spikes"]
        X_stim = rec[:, 0 : self.input_len]
        X_spikes = spikes[0 : self.input_len]
        X = np.vstack((X_stim, X_spikes))
        Y = spikes[self.input_len :]
        assert Y.shape == (self.output_len,)
        y = np.sum(Y)
        return X, y


class DistFieldDataset(torch.utils.data.Dataset):
    """
    Dataset producing a snippet (spikes and stimulus) and a spike distance
    field. A portion of the spikes in the snippet are masked. The spike distance
    field will be created for the masked portion of the snippet.

    The intended usecase of this dataset is to predict spike activity given
    stimulus and spike history.
    """

    # Mask value should be negative. Zero represents no spikes, and 1+ represent
    # a spike count which can be created than 1!
    MASK_VALUE = -1
    # Do not! set a seed within the dataset. Process forking leads to identical
    # seeds.
    # RNG_SEED = 123

    # TODO: make configurable
    NOISE_SD = 0.1
    NOISE_MU = 0
    NOISE_JITTER = 4
    DROP_RATE = 0.1

    def __init__(
        self,
        recording: mea.SpikeRecording,
        snippet_len,
        mask_begin: int,
        mask_end: int,
        pad: int,
        dist_clamp: float,
        stride: int = 1,
        enable_augmentation: bool = True,
        allow_cheating: bool = False,
    ):
        self.enable_augmentation = enable_augmentation
        self.num_stim_channels = recording.stimulus.shape[1]
        self.pad = pad
        self.dist_clamp = dist_clamp
        self.ds = SnippetDataset(
            recording, snippet_len + self.pad, stride=stride
        )
        self.mask_slice = slice(mask_begin, mask_end)
        self.allow_cheating = allow_cheating
        self.stim_mean = np.expand_dims(recording.stimulus.mean(axis=0), -1)
        self.stim_sd = np.expand_dims(recording.stimulus.std(axis=0), -1)
        # The stimulus mean will be dominated by the mask
        mask_len = mask_end - mask_begin
        self.spike_mean = mask_len * self.MASK_VALUE / snippet_len
        self.spike_sd = (
            mask_len * (self.MASK_VALUE - self.spike_mean) ** 2
            + (snippet_len - mask_len) * self.spike_mean**2
        ) / snippet_len

    def __len__(self):
        return len(self.ds)

    @property
    def recording(self):
        return self.ds.recording

    @property
    def sample_rate(self):
        return self.recording.sample_rate

    @property
    def stride(self):
        return self.ds.stride

    @classmethod
    def mask_start(cls, spikes):
        mask_val = cls.MASK_VALUE
        mask_start_idx = np.min(np.flatnonzero(spikes == mask_val))
        return mask_start_idx

    def _augment_stimulus(self, stimulus):
        """
        Augment a stimulus portion of a sample.
        """
        # Whole block scale.
        mu = 1.0
        sd = 0.10
        scale = np.random.normal(mu, sd, size=(1,))
        # Whole block offset.
        mu = 0.0
        sigma = 0.10
        offset_noise = np.random.normal(mu, sigma, size=(1,))
        # Per bin noise.
        max_length = stimulus.shape[1]
        center, length = np.random.randint(low=0, high=max_length, size=(2,))
        left = max(0, center - length // 2)
        right = min(max_length - 1, center + length // 2 + 1)
        bin_noise = np.random.normal(
            self.NOISE_MU,
            self.NOISE_SD,
            size=(self.num_stim_channels, (right - left)),
        )
        stimulus = stimulus * scale + offset_noise
        stimulus[:, left:right] += bin_noise
        return stimulus

    def normalize_stimulus(self, stimulus):
        """
        Normalize a stimulus portion of a sample.
        """
        res = (stimulus - self.stim_mean) / self.stim_sd
        return res

    def normalize_spikes(self, spikes):
        """
        Normalize a spike portion of a sample.
        """
        res = (spikes - self.spike_mean) / self.spike_sd
        return res

    def normalize_snippet(self, snippet):
        snippet[0 : self.num_stim_channels, :] = (
            snippet[0 : self.num_stim_channels, :] - self.stim_mean
        ) / self.stim_sd
        snippet[self.num_stim_channels :, :] = (
            snippet[self.num_stim_channels :] - self.spike_mean
        ) / self.spike_sd
        return snippet

    def _augment_spikes(self, spikes):
        """
        "Augment the spike portion of a sample.

        Call this on the model input portion of the spike data, and not the
        portion that we are trying to predict.
        """
        spike_indicies = np.nonzero(spikes)
        spikes[spike_indicies] = 0
        # Add jitter
        jitter = np.random.randint(
            -self.NOISE_JITTER, self.NOISE_JITTER, len(spike_indicies)
        )
        spike_indicies = np.clip(spike_indicies + jitter, 0, len(spikes) - 1)
        # Drop some spikes.
        new_spikes = np.random.binomial(
            1, p=(1 - self.DROP_RATE), size=len(spike_indicies)
        )
        spikes[spike_indicies] = new_spikes
        return spikes

    def __getitem__(self, idx):
        """
        Returns the sample at the given index.

        +---------------------------------+-------------+
        |  a) input stimulus                            |
        +---------------------------------+-------------+---------+
        |  b) input spike data            | c) masked   | d) pad* |
        +---------------------------------+-------------+---------+

        Note (d*): there is an extra bit of spike data used when creating
        a sample, here called a pad. The pad is used to calculate the ground
        truth distance field. This bit of data is not placed in the sample that
        is returned.
        """
        # 1. Get the snippet. Make it extra long, for the distance field calc.
        sample = self.ds[idx]
        extra_long_stimulus = sample["stimulus"]
        extra_long_spikes = sample["spikes"]
        cluster_id = sample["cluster_id"]
        # For some unknown reason, the following copy call makes
        # training about 5x faster, and it has no effect when called on the
        # stimulus array. Maybe related to the copy that is done below for
        # target_spikes?
        extra_long_spikes = np.array(extra_long_spikes, copy=True, dtype=float)

        # 2. Optional augmentation.
        if self.enable_augmentation:
            extra_long_stimulus = self._augment_stimulus(extra_long_stimulus)
            extra_long_spikes[0 : self.mask_slice.start] = self._augment_spikes(
                extra_long_spikes[0 : self.mask_slice.start]
            )
        # 3. Calculate the distance field.
        dist = sdf.distance_field(extra_long_spikes, self.dist_clamp)
        # With the distance field calculated, we can throw away the extra bit.
        dist = dist[self.mask_slice]
        target_spikes = np.array(extra_long_spikes[self.mask_slice], copy=True)
        if not self.allow_cheating:
            extra_long_spikes[self.mask_slice] = self.MASK_VALUE
        # 4. Remove the extra padding that was used to calculate the distance fields.
        stimulus = extra_long_stimulus[:, 0 : -self.pad]
        spikes = extra_long_spikes[0 : -self.pad]
        # 5. Normalize
        stimulus_norm = self.normalize_stimulus(stimulus)
        # TODO: what about cheating? Why now?
        # 6. Stack 
        snippet = np.vstack((stimulus_norm, spikes))
        # Returning a dictionary is more flexible than returning a tuple, as
        # we can add to the dictionary without breaking existing consumers of
        # the dataset.
        res = {
            "snippet": snippet,
            "dist": dist,
            "target_spikes": target_spikes,
            "cluster_id": cluster_id,
        }
        return res



class LabeledConcatDataset(torch.utils.data.Dataset):
    """
    Dataset that concatenates datasets and inserts a dataset label.

    This is an edited version of PyTorch's ConcatDataset, with the addition
    of a dataset index being included in each sample. This is useful for
    making a multi-recording dataset easily from a list of single recording
    datasets. The PyTorch implementation wasn't sufficient, as we want to
    include information of which recording a sample belongs to.
    """

    datasets: List[torch.utils.data.Dataset]
    cumulative_sizes: List[int]

    @staticmethod
    def cumsum(sequence):
        r = []
        s = 0
        for e in sequence:
            l = len(e)
            r.append(l + s)
            s += l
        return r

    def __init__(
        self,
        datasets: Iterable[torch.utils.data.Dataset],
        label_key: str = "id",
    ) -> None:
        super(LabeledConcatDataset, self).__init__()
        self.datasets = list(datasets)
        self.label_key = label_key
        assert len(self.datasets) > 0, (
            "datasets should not be an empty " "iterable"
        )
        for d in self.datasets:
            assert not isinstance(
                d, torch.utils.data.IterableDataset
            ), "ConcatDataset does not support IterableDataset"
        self.cumulative_sizes = self.cumsum(self.datasets)

    def __len__(self):
        return self.cumulative_sizes[-1]

    def __getitem__(self, idx):
        if idx < 0:
            if -idx > len(self):
                raise ValueError(
                    "absolute value of index should not exceed dataset length"
                )
            idx = len(self) + idx
        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
        sample = self.datasets[dataset_idx][sample_idx]
        assert isinstance(
            sample, dict
        ), "Sample must be a dictionary in order to add a label."
        sample[self.label_key] = dataset_idx
        return sample


class ConcatDistFieldDataset(LabeledConcatDataset):
    """
    A concatenation of distfield datasets, one for each recording.

    This class exists to expose some methods like sample_rate() that are both
    present and consistent among all contained datasets. Using these methods
    is tedious if using LabeledConcatDataset directly. It's also brittle, as
    the LabeledConcatDataset can't (but should) act as a drop in replacement
    for a single DistFieldDataset.

    While we are at it, we can encapsulate the setting of the "rec_id" as the
    custom label key.
    """

    def __init__(self, datasets: Iterable[DistFieldDataset]) -> None:
        super().__init__(datasets, label_key="rec_id")
        # Check for consistency
        sample_rate = self.datasets[0].sample_rate
        for ds in self.datasets:
            assert ds.sample_rate == sample_rate, (
                "All datasets must have the same sample rate. Got "
                f"({sample_rate}) and ({ds.sample_rate})."
            )

    @property
    def sample_rate(self):
        return self.datasets[0].sample_rate

    @property
    def stride(self):
        return self.datasets[0].stride




import pytest
import retinapy.dataset as dataset
import retinapy.mea as mea
import numpy as np
import pathlib


DATA_DIR = pathlib.Path("./data/ff_noise_recordings")


@pytest.fixture
def exp12_1kHz():
    rec = mea.single_3brain_recording("Chicken_17_08_21_Phase_00", DATA_DIR)
    rec = mea.decompress_recording(rec, downsample=18)
    return rec


def test_DistFieldDataset(exp12_1kHz):
    # Setup
    snippet_len = 2000
    mask_begin = 1000
    mask_end = 1500
    max_dist = 500
    pad = 500
    mask_shape = (mask_end - mask_begin,)

    # Test
    # 1. The dataset should be created correctly.
    exp12_1kHz_cluster_13 = exp12_1kHz.clusters({13})
    ds = dataset.DistFieldDataset(
        exp12_1kHz_cluster_13, snippet_len, mask_begin, mask_end, pad, max_dist
    )
    # 2. The dataset should have the correct length.
    assert len(ds) == len(exp12_1kHz) - snippet_len - pad + 1

    # 3. The dataset should return the correct snippet and distance fields.
    sample = ds[0]
    masked_snippet = sample["snippet"]
    target_spikes = sample["target_spikes"]
    dist_field = sample["dist"]
    # 3.1. The shapes should be correct.
    # The first dimension: LEDs (4) and spikes (1).
    assert masked_snippet.shape == (mea.NUM_STIMULUS_LEDS + 1, snippet_len)
    assert target_spikes.shape == mask_shape
    assert dist_field.shape == mask_shape
    # 3.2 The target spikes should be float array, with mostly zeros.
    # Note: this was changed from int. float resulted in higher examples/sec
    # when training.
    assert target_spikes.dtype == float
    known_spike_count = 10
    assert np.sum(target_spikes) == known_spike_count
    # 3.3 The distance fields should be zero where there are spikes, and
    #     non-zero where there are no spikes.
    assert not np.any(dist_field[np.where(target_spikes > 0)])
    assert np.all(dist_field[np.where(target_spikes == 0)])
    # 3.4 No distance in the distance fields should be larger than max_dist.
    assert np.max(dist_field) <= max_dist


@pytest.fixture
def two_exps():
    recs = mea.load_3brain_recordings(
        DATA_DIR,
        include=["Chicken_04_08_21_Phase_01", "Chicken_20_08_21_Phase_00"],
    )
    dc_recs = mea.decompress_recordings(recs, downsample=18)
    return dc_recs


@pytest.mark.parametrize("stride", [1, 3, 17])
def test_ConcatDistFieldDataset(two_exps, np_rng, stride):
    """Test the concatenated distfield dataset.

    Tests that:
        1. Construction from two DistField datasets throws no errors and has 
            expected length.
        2. Samples with same timestep but different cluster have equal stimulus.
        3. All samples for a single cluster have the correct (and same) 
            "cluster_id" and "rec_id" value.

    These tests are run parameterized for different strides.
    """
    # Setup
    num_timestep_trials = 1000
    num_cluster_trials = 50
    snippet_len = 1000
    mask_begin = 800
    mask_end = 1000
    pad = 100
    dist_clamp = 500
    distfield_ds1 = dataset.DistFieldDataset(
        two_exps[0],
        snippet_len,
        mask_begin,
        mask_end,
        pad,
        dist_clamp,
        stride=stride,
        enable_augmentation=False,
    )
    distfield_ds2 = dataset.DistFieldDataset(
        two_exps[1],
        snippet_len,
        mask_begin,
        mask_end,
        pad,
        dist_clamp,
        stride=stride,
        enable_augmentation=False,
    )

    # Test
    # 1. The dataset should be created correctly.
    concat_ds = dataset.ConcatDistFieldDataset([distfield_ds1, distfield_ds2])
    assert len(concat_ds) == len(distfield_ds1) + len(distfield_ds2)

    # 2. Two clusters from the same recording should share the same stimulus.
    # 2.1 For the first recording.
    num_timesteps1 = distfield_ds1.ds.num_strided_timesteps
    test_idxs = np_rng.integers(0, num_timesteps1, num_timestep_trials)
    for idx in test_idxs:
        s1 = concat_ds[idx]["snippet"][0 : mea.NUM_STIMULUS_LEDS, :]
        s2 = concat_ds[idx + num_timesteps1]["snippet"][
            0 : mea.NUM_STIMULUS_LEDS, :
        ]
        np.testing.assert_allclose(s1, s2, err_msg=f"idx={idx}")
    # 2.2 For the second recording.
    num_timesteps2 = distfield_ds2.ds.num_strided_timesteps
    ds_2_start_idx = len(distfield_ds1)
    test_idxs = np_rng.integers(
        ds_2_start_idx, ds_2_start_idx + num_timesteps2, num_timestep_trials
    )
    for idx in test_idxs:
        s1 = concat_ds[idx]["snippet"][0 : mea.NUM_STIMULUS_LEDS, :]
        s2 = concat_ds[idx + num_timesteps2]["snippet"][
            0 : mea.NUM_STIMULUS_LEDS, :
        ]
        np.testing.assert_allclose(s1, s2, err_msg=f"idx={idx}")

    # 3. A cluster's snippets should have the same 'rec_id' and 'cluster_id'.
    # 3.1 For the first recording.
    cluster_idxs = np_rng.integers(
        0, len(distfield_ds1.recording.cluster_ids), num_cluster_trials
    )
    for c_idx in cluster_idxs:
        offset = c_idx * distfield_ds1.ds.num_strided_timesteps
        test_idxs = np_rng.integers(0, num_timesteps1, num_timestep_trials)
        for idx in test_idxs:
            c_idx_from_ds = concat_ds[idx + offset]["cluster_id"]
            rec_idx_from_ds = concat_ds[idx + offset]["rec_id"]
            assert c_idx_from_ds == c_idx
            assert rec_idx_from_ds == 0
    # 3.2 For the second recording.
    cluster_idxs = np_rng.integers(
        0, len(distfield_ds2.recording.cluster_ids), num_cluster_trials
    )
    for c_idx in cluster_idxs:
        offset = (
            len(distfield_ds1) + c_idx * distfield_ds2.ds.num_strided_timesteps
        )
        test_idxs = np_rng.integers(0, num_timesteps2, num_timestep_trials)
        for idx in test_idxs:
            c_idx_from_ds = concat_ds[idx + offset]["cluster_id"]
            rec_idx_from_ds = concat_ds[idx + offset]["rec_id"]
            assert c_idx_from_ds == c_idx
            assert rec_idx_from_ds == 1



import numpy as np
import math
from typing import Optional, OrderedDict, Tuple
from collections import namedtuple
from collections import defaultdict
import torch
import torch.nn.functional as F

"""
Notes:

    - Having a single array that uses +ve and -ve values to signal the direction
      towards the closest spike would be an implementation more close to the
      signed distance function approach. However, there is an issue where a
      timestep might be in the middle of two very distant spikes. In this case,
      we do not want there to be a large penalty for predicting say -50 instead
      of +50, where this vacillation might appear if there is uncertainty
      as to whether the timestep is slightly closer to the earlier or later 
      spike.
    - To avoid the above problem, have the model output both a forward distance
      *and* a reverse distance.
    - [A while later...] The bi-directional approach produces a loss function 
      that value jumps that are too large and abrupt. So, going back to the 
      original idea. A second way to address the concern above is to make it a
      non-signed distance. That's what I'll try next (and probably should have 
      tried first).
"""

def bi_distance_field(spikes: np.ndarray, default_dist: int, 
                      max_dist: Optional[int]=None):
    """
    Calculates the bi-directional distance field of a spike train.

    Two distances are calculated for each timestep:
        - how far the timestep is since the previous spike ("distance after")
        - how far the timestep is since the next spike ("distance before")

    For example, if a spike happens at timestep t=4, then we have:
        a) dist_after[4] = dist_before[4] = 0
        b) dist_after[5] = 1
        c) dist_before[3] = 1
        d) we can't say anything about dist_after[3] or dist_before[5].

    
    The distance field is initialized with the default_dist value.
    max_dist sets the maximum the maximum distance value.
    """
    raise NotImplementedError("Failing test. See GitHub issue #2.")
    dist_after = np.full_like(spikes, default_dist, int)
    dist_before = np.full_like(spikes, default_dist, int)
    spike_indicies = (spikes == 1).nonzero()[0]
    r = np.arange(0, len(spikes))
    if spike_indicies is not None:
        # after
        endpoints = list(spike_indicies)
        diff = np.diff(endpoints)
        for idx in range(len(endpoints) - 1):
            dist_after[endpoints[idx] : endpoints[idx + 1]] = r[0 : diff[idx]]
        # before
        endpoints = [-1] + list(spike_indicies)
        diff = np.diff(endpoints)
        for idx in range(len(endpoints) - 1):
            dist_before[endpoints[idx] + 1 : endpoints[idx + 1] + 1] = np.flip(
                r[0 : diff[idx]]
            )
    if max_dist:
        dist_before = np.minimum(dist_before, max_dist)
        dist_after = np.minimum(dist_after, max_dist)
    return dist_before, dist_after


def distance_field(spikes: np.ndarray, default_distance: float):
    """
    Calculates the distance field of a spike train.

    Args:
        spikes: a 1D array where a 1 represents a timestep where a spike
            occurred, and 1 where a spike did not occur.
        default_distance: the value to initialize each element of the
            distance field. This functions as a maximum distance. Notably, if
            there are no spikes in the spikes array, then all elements of the
            distance field will be set to this value.
    """
    dist_field = np.full_like(spikes, default_distance, float)
    spike_indicies = (spikes == 1).nonzero()[0]
    all_indicies = np.arange(len(spikes))
    for s in spike_indicies:
        dist_field = np.minimum(dist_field, np.abs(all_indicies - s))
    return dist_field


def spike_interval(spikes: np.ndarray, default_count: int):
    """Count the timesteps between spikes"""
    count_field = np.full_like(spikes, default_count, int)
    spike_indicies = (spikes == 1).nonzero()[0]
    count_field[spike_indicies] = 0
    counts = np.diff(spike_indicies) - 1
    for idx in range(len(counts)):
        count_field[spike_indicies[idx] + 1 : spike_indicies[idx + 1]] = counts[idx]
    return count_field


def distance_field2(spikes, default_dist):
    """An alternative (non-vector) distance field implementation.
    
    Not used at the moment. Leaving it here for reference. 
    """
    dist = [default_dist,] * len(spikes)
    spike_indicies = [idx for idx,v in enumerate(spikes) if v == 1]

    def _dfs(idx, cur_dist):
        if dist[idx] <= cur_dist:
            return
        dist[idx] = cur_dist
        if idx > 0:
            _dfs(idx - 1, cur_dist + 1)
        if idx < len(spikes) - 1:
            _dfs(idx + 1, cur_dist + 1)

    for s in spike_indicies:
        _dfs(s, cur_dist=0)
    return dist


# Below are various attempts at inference. Very WIP.

def quick_inference_from_df(dist, target_interval, threshold=0.1):
    dist = dist[:, target_interval[0] : target_interval[1]]
    num_spikes = torch.sum(dist < threshold, dim=1)
    return num_spikes


def quick_inference_from_df2(dist, target_interval, threshold=0.1):
    dist = dist[:, target_interval[0] : target_interval[1]]
    kernel = torch.FloatTensor(
        [[[0.006, 0.061, 0.242, 0.383, 0.242, 0.061, 0.006]]]
    ).to(dist.device)
    smoothed = torch.squeeze(F.conv1d(torch.unsqueeze(dist, 1), kernel))
    below_threshold = (smoothed < threshold).float()
    # below_threshold = below_threshold[:,::3]
    transitions = torch.sum((torch.diff(below_threshold, 1, dim=1) > 0), dim=1)
    return transitions


def quick_inference_from_bi_df(
    dist_before, dist_after, target_interval, threshold=0.1
):
    dist_before = dist_before[:, target_interval[0] : target_interval[1]]
    dist_after = dist_after[:, target_interval[0] : target_interval[1]]
    # match = torch.sum((dist_after < threshold) * (dist_before < threshold), dim=1)
    match = torch.sum((dist_after < threshold), dim=1)
    return match


def mle_inference_from_df(
    dist: torch.Tensor,
    lhs_spike,
    rhs_spike,
    spike_pad,
    max_clamp=None,
    max_num_spikes=100,
    resolution=1,
):
    if len(dist.shape) != 1:
        raise ValueError("Batching isn't supported (yet)")
    init_a = max(0, lhs_spike + spike_pad + 1)
    init_b = min(len(dist) - 1, rhs_spike - spike_pad - 1)
    max_n = int(math.ceil((init_b - init_a) / (spike_pad + 1)))
    max_n = min(max_n, max_num_spikes)
    _len = len(dist)
    device = dist.device

    # If a-1 and b+1 are the indicies of two spikes, what is the energy 
    # contributed by the elements in (a,b)?
    memo = {}  # (a,b, num_allowed) -> ('energy')

    zero_spike_memo = torch.zeros(_len, _len, device=device)
    for i in range(_len):
        for j in range(i, _len):
            d_after = torch.arange(
                j - i + 1, dtype=torch.float32, device=device
            )
            d_before = torch.flip(
                torch.arange(j - i + 1, dtype=torch.float32, device=device),
                dims=(0,),
            )
            # The endpoints need special treatment.
            if i == 0:
                d_after += -lhs_spike
            if j == 0:
                d_before += rhs_spike
            d_min = torch.clamp(torch.minimum(d_after, d_before), max=max_clamp)
            energy = torch.sum(torch.abs(dist[i : j + 1] - d_min))
            zero_spike_memo[i, j] = energy

    global_best_energy = math.inf
    low_energy_positions = dist < 20

    def _dfs(a, b, energy_so_far, num_allowed_spikes) -> Tuple[float, Tuple[int, ...]]:
        nonlocal global_best_energy
        if a >= b:
            return 0, ()
        if (a, b, num_allowed_spikes) in memo:
            return memo[(a, b, num_allowed_spikes)]
        if energy_so_far > global_best_energy:
            return math.inf, ()
        no_spike_energy = zero_spike_memo[a, b]
        best_energy = no_spike_energy
        best_seq = ()
        if not num_allowed_spikes:
            return best_energy, best_seq
        for candidate_pos in range(a, b + 1, resolution):
            if not low_energy_positions[candidate_pos]:
                continue
            for num_l_spikes in range(num_allowed_spikes):
                for num_r_spikes in range(num_allowed_spikes - num_l_spikes):
                    lhs_energy, lhs_seq = _dfs(a, candidate_pos - 1, energy_so_far, num_l_spikes)
                    spike_pos_energy = min(dist[candidate_pos], max_clamp)
                    energy = energy_so_far + lhs_energy + spike_pos_energy
                    rhs_energy, rhs_seq = _dfs(candidate_pos + 1, b, energy, num_r_spikes)
                    energy += rhs_energy
                    if energy < best_energy:
                        best_energy = energy
                        best_seq = lhs_seq + (candidate_pos,) + rhs_seq
                        if (best_energy+energy_so_far) < global_best_energy:
                            global_best_energy = best_energy
        memo[(a, b)] = (best_energy, best_seq, num_allowed_spikes)
        return best_energy, best_seq

    e, seq = _dfs(0, _len - 1, energy_so_far=0, num_allowed_spikes=max_n)
    return e, seq


def count_inference_from_bi_df2(
    dist_before,
    dist_after,
    lhs_spike,
    rhs_spike,
    spike_pad,
    target_interval,
    max_num_spikes=100,
    resolution=1,
):
    """
    Target_interval will be (start, end), where we will consider spikes to
    be allowed at any index from start to (end-1) inclusive (ie right open
    interval [start, end) ).
    """
    raise NotImplementedError("TODO: fix broken tests. GitHub issue #1.")
    if dist_before.shape != dist_after.shape:
        raise ValueError("dist_before and dist_after must be the same shape.")
    if torch.any(lhs_spike >= 0):
        raise ValueError(
            "The left hand side spike should be -ve steps " "from interval."
        )
    device = dist_before.device
    _len = dist_before.shape[-1]
    batch_len = dist_before.shape[0]
    max_n = int(math.ceil(_len / (spike_pad + 1)))
    # Memo of energy of intervals between two spikes.
    after_memo = torch.zeros((batch_len, _len, _len), device=device)
    before_memo = torch.zeros((batch_len, _len, _len), device=device)
    for i in range(_len):
        for j in range(i, _len):
            # i,j represent the start and end of the interval *between* the
            # spikes. So the spikes happen at i-1 and j+1. This means that
            # i=j is valid.
            # E.g. if (i,j) == (2,4), then result is dist((1,2,3,4), dist_after)
            # TODO: cache these distances/ranges?
            # Need to make the dtype a float, as cdist requires float.
            d_after = torch.arange(
                j - i + 1, dtype=torch.float32, device=device
            ).repeat(batch_len, 1)
            d_before = torch.flip(
                torch.arange(j - i + 1, dtype=torch.float32, device=device),
                dims=(0,),
            ).repeat(batch_len, 1)
            # The endpoints need special treatment.
            if i == 0:
                d_after += -lhs_spike
            if j == 0:
                d_before += rhs_spike
            energy_after = (dist_after[:, i : j + 1] - d_after).pow(2).sum(1)
            energy_before = (dist_before[:, i : j + 1] - d_before).pow(2).sum(1)
            after_memo[:, i, j] = torch.squeeze(energy_after)
            before_memo[:, i, j] = torch.squeeze(energy_before)

    # At this point, it's only valid to query the memo.
    # Make a list of every possible spike sequence.
    spike_sequences = []

    def _dfs(a, spikes: Tuple[int, ...]):
        if a >= _len or len(spikes) >= max_num_spikes:
            # The end of the tree has been reached.
            spike_sequences.append(spikes)
            return
        # Either there is no spike here:
        _dfs(a + 1 + (resolution - 1), spikes)
        # Or there is as spike.
        next_a = a + spike_pad + 1
        _dfs(next_a, spikes + (a,))

    _dfs(0, spikes=())

    # Now get the energy of every possible spike sequence. Bin by the number
    # of spikes in the target interval, as that's why we are here in the
    # first place.
    by_n = defaultdict(list)
    for ss in spike_sequences:
        interval_count = 0
        for i in range(len(ss)):
            # TODO: check, are we off by one here?
            if ss[i] < target_interval[0]:
                # Interval hasn't started yet.
                continue
            # TODO: check, are we off by one here?
            elif ss[i] >= target_interval[1]:
                # Gone past interval.
                break
            interval_count += 1
        # The energy is the sum of:
        #   - the energy all steps where there is a spike
        #   - the energy of all spike-empty intervals thus formed.
        # Before and after energies will be accumulated here.
        # The energies are added, which corresponds to multiplication of
        # probabilities. We could just have one energy tally, but keeping
        # it as two for debugging.
        after_energy = torch.zeros((batch_len,), device=device)
        before_energy = torch.zeros((batch_len,), device=device)
        prev_spike = -1
        for s_idx in range(len(ss)):
            cur_spike = ss[s_idx]
            if cur_spike > prev_spike + 1:
                # There is a spike-empty interval to consider.
                after_energy += after_memo[:, prev_spike + 1, cur_spike]
                before_energy += after_memo[:, prev_spike + 1, cur_spike]
            # Don't do this:
            #    if ss[s_idx + 1] > cur_spike + 1:
            # Don't double count. The next iteration will get the other side.
            # Add the energy of the spike timestep.
            after_energy += dist_after[:, cur_spike]
            before_energy += dist_before[:, cur_spike]
            prev_spike = cur_spike
        # The last spike potentially has an spike-empty interval on the rhs.
        # This if-block also handles the case when ss = (,)
        if prev_spike < (_len - 1):
            after_energy += after_memo[:, prev_spike + 1, _len - 1]
            before_energy += after_memo[:, prev_spike + 1, _len - 1]
        by_n[interval_count].append((before_energy + after_energy))

    # We are almost there. Just find the n with the maximum energy.
    # The energies for each n are currently saved as a list of energies for
    # independent events, so we need to combine them into one value. This
    # is like adding probabilities, which must be done in exponent space
    # (log space).
    total_by_n = torch.zeros((batch_len, max_n + 1), device=device)
    for n in range(max_n + 1):
        if len(by_n[n]):
            total_by_n[:, n] = torch.logsumexp(
                torch.stack(by_n[n], dim=1), dim=1
            )
        else:
            total_by_n[:, n] = 0
    # The actual answer is obtained by a simple argmax.
    res = torch.argmax(total_by_n, dim=1)
    return res


def count_inference_from_bi_df(
    dist_before,
    dist_after,
    lhs_spike,
    rhs_spike,
    spike_pad,
    target_interval,
    max_num_spikes,
):
    raise NotImplementedError("TODO: fix broken tests. GitHub issue #1.")
    if len(dist_before) != len(dist_after):
        raise ValueError("dist_before and dist_after must be the same length.")
    # A 1 element query must have r-l >= 2.
    # l r
    # 012
    # |-|
    num_elements = rhs_spike - lhs_spike - 1
    if num_elements <= 0:
        raise ValueError(
            "Invalid query range ((lhs, rhs) = ({lhs_spike}, {rhs_spike}))."
        )
    if target_interval[0] < 0 or target_interval[1] > len(dist_before):
        raise ValueError(
            f"The query interval ({target_interval}) must be "
            "within the range of the distance field "
            f"(len: {len(dist_before)})."
        )

    init_a = max(0, lhs_spike + spike_pad + 1)
    init_b = min(len(dist_before) - 1, rhs_spike - spike_pad - 1)
    max_n = int(math.ceil((init_b - init_a) / (spike_pad + 1)))
    max_n = min(max_n, max_num_spikes)
    target_energy_by_n = torch.zeros(
        (max_n + 1),
    )  # np.zeros((max_n + 1,))

    memo = {}
    # range_cache = np.arange(0, rhs_spike - lhs_spike)
    range_cache = torch.arange(
        0, rhs_spike - lhs_spike, device=dist_before.device,
    )
    MemoKey = namedtuple("MemoKey", ["l_spike", "r_spike", "a", "b", "n"])

    def energy_to_prob(e):
        b = 1  # / len(dist_before)
        return math.exp(b * -e)

    def _energy(l_spike, r_spike, a, b, n):
        print(f"l_spike: {l_spike}, r_spike: {r_spike}, a: {a}, b: {b}, n: {n}")
        assert l_spike < a <= b < r_spike
        assert n >= 0
        key = MemoKey(l_spike, r_spike, a, b, n)
        if key in memo:
            # print('hit')
            return memo[key]
        ans = []
        if n == 0:
            dist_after_measured = range_cache[(a - l_spike) : (b - l_spike + 1)]
            dist_before_measured = torch.flip(  # np.flip(
                range_cache[(r_spike - b) : (r_spike - a) + 1], dims=(0,)
            )
            assert len(dist_after_measured) == len(dist_before_measured)
            # dist_after_measured = torch.clip(dist_after_measured, 0, max_dist)
            # dist_before_measured = torch.clip(dist_before_measured, 0, max_dist)
            energy_forward = torch.sum(  # np.sum(
                (dist_before[a : b + 1] - dist_before_measured) ** 2
            )
            energy_backward = torch.sum(  # np.sum(
                (dist_after[a : b + 1] - dist_after_measured) ** 2
            )
            ans.append(energy_forward + energy_backward)
            # energy_to_prob(energy_forward + energy_backward)
        else:
            assert n > 0
            start = max(a, l_spike + spike_pad + 1)
            # end *includes* the last valid index.
            # with two spikes needing to be inserted, we need at least
            # 2*(spike_pad+1) from the end.
            end = min(b, r_spike - n * (spike_pad + 1))
            if start > end:
                return 0
            for x in range(start, end + 1):
                # energy_at_x = min(dist_after[x], max_dist) + min(
                #    dist_before[x], max_dist
                # )
                energy_at_x = dist_after[x] + dist_before[x]
                # prob_at_x = energy_to_prob(energy_at_x)
                # prob_rhs = prob_lhs = 1
                energy_rhs = energy_lhs = 0
                if x > a:
                    energy_lhs = _energy(l_spike, x, a, x - 1, n=0)
                if x < b:
                    energy_rhs = _energy(x, r_spike, x + 1, b, n=n - 1)
                if x == b and n > 1:
                    # Zero liklihood if there are too many n left to fit on rhs.
                    energy_rhs = 0
                event_energy = energy_at_x + energy_lhs + energy_rhs
                ans.append(event_energy)
        total_energy = torch.logsumexp(torch.Tensor(ans), dim=0)
        # Save result
        memo[key] = total_energy
        # Check against target interval.
        if a == target_interval[0] and b == target_interval[1] - 1:
            target_energy_by_n[n] += ans
        return ans

    last_p = 0
    for i in range(max_n + 1):
        p = _energy(lhs_spike, rhs_spike, init_a, init_b, i)
        print(f"i: {i}:\t p: {p}")
        if p < last_p:
            break
        last_p = p

    return torch.argmax(target_energy_by_n)  # np.argmax(target_energy_by_n)

